import numpy as np
import autograd.numpy as anp
import random
import warnings
from .abstract_function import AbstractFunction
from .constant import Constant
from .affine_form import AffineForm
from .variable import Variable
from .utilities import check_input


class Max(AbstractFunction):
    """Max function (elementwise)

    .. math::

        f(x,y) = \\max(x,y)

    with :math:`x,y: \\mathbb{R}^{n}`.

    Args:
        f1 (AbstractFunction): input function
        f2 (AbstractFunction): input function

    Raises:
        TypeError: input must be a AbstractFunction object
        ValueError: sunctions must have the same input/output shapes
    """

    def __init__(self, f1: AbstractFunction, f2: AbstractFunction):
        if not (isinstance(f1, AbstractFunction) or isinstance(f2, AbstractFunction)):
            raise TypeError("At least one input must be a AbstractFunction object, otherwise use the builtin max operator.")
        
        self.f1 = f1
        self.f2 = f2

        if not isinstance(f1, AbstractFunction):
            if isinstance(f1, (np.ndarray, float, int)):
                v = Variable(f2.input_shape[0])
                A = np.zeros([f2.input_shape[0], f2.output_shape[0]])
                if isinstance(f1, np.ndarray):
                    if f1.shape[1] != 1:
                        f1 = f1.transpose()
                self.f1 = A @ v + f1
            else: 
                raise TypeError("Only AbstractFunction, numpy.ndarray, float and int are supported")
        
        if not isinstance(f2, AbstractFunction):
            if isinstance(f2, (np.ndarray, float, int)):
                v = Variable(f1.input_shape[0])
                A = np.zeros([f1.input_shape[0], f1.output_shape[0]])
                if isinstance(f2, np.ndarray):
                    if f2.shape[1] != 1:
                        f2 = f2.transpose()
                self.f2 = A @ v + f2
            else:
                raise TypeError("Only AbstractFunction, numpy.ndarray, float and int are supported")

        if self.f1.input_shape != self.f2.input_shape or self.f1.output_shape != self.f2.output_shape:
            raise ValueError("Different input/output shapes")
        
        if not (self.f1.is_differentiable and self.f2.is_differentiable):
            warnings.warn(
                'Composition with a nondifferentiable function will lead to an\
                    error when asking for a subgradient')

        self.input_shape = self.f1.input_shape
        self.output_shape = self.f1.output_shape

        self.affine = False
        self.quadratic = False

        super().__init__()

    def _expression(self):
        expression = 'Max({}, {})'.format(self.f1._expression(), self.f2._expression())
        return expression 
    
    def _to_cvxpy(self):
        import cvxpy as cvx
        return cvx.maximum(self.f1._to_cvxpy(), self.f2._to_cvxpy())
    
    def _extend_variable(self, n_var, axis, pos):
        return Max(self.f1._extend_variable(n_var, axis, pos), self.f2._extend_variable(n_var, axis, pos))

    @check_input
    def eval(self, x: np.ndarray) -> np.ndarray:
        return ((self.f1.eval(x) + self.f2.eval(x) + anp.abs(self.f1.eval(x) - self.f2.eval(x)))/2).reshape(self.output_shape)

    # @check_input
    # def jacobian(self, x: np.ndarray, **kwargs) -> np.ndarray:
    #     if not (self.f1.is_differentiable and self.f2.is_differentiable):
    #         warnings.warn("Composition of nondifferentiable functions. The Jacobian may be not correct.")
    #     # else:
    #     j1 = self.f1.jacobian(x, **kwargs)
    #     j2 = self.f2.jacobian(x, **kwargs)
    #     diff = self.f1.eval(x) - self.f2.eval(x)
    #     for elem in range(diff.size):
    #         if diff[elem] == 0:
    #             diff[elem] = random.uniform(-1, 1)
    #     res = j1 + j2 + np.diag(np.sign(diff).flatten()) @ (j1 - j2)
    #     return res/2



import numpy as np
from .linear_problem import LinearProblem
from ..functions import AffineForm
from ..functions.affine_form import aggregate_affine_form
from ..constraints import Constraint, AbstractSet
from cvxopt.glpk import ilp
from cvxopt import matrix


class MixedIntegerLinearProblem(LinearProblem):
    """Solve a Mixed-Integer Linear programming problem defined as:

    .. math::

        \\text{minimize } & c^\\top x

        \\text{subject to } & G x \\leq h

                            & A x = b

                            & x_k \\in \\mathbb{Z}, \\forall k \\in I

                            & x_k \\in \\{0,1\\}, \\forall k \\in B
        
    where :math:`I` is the set of integer variable indexes 
    and :math:`B` is the set of binary variable indexes.
    """

    def __init__(self, objective_function: AffineForm, constraints: list = None, integer_vars: list = None, binary_vars: list = None, **kwargs):
        """[summary]

        Args:
            objective_function (AffineForm): [description]
            constraints (list, optional): [description]. Defaults to None.
            integer_vars (list, optional)
            binary_vars (list, optional)
        """
        super().__init__(objective_function=objective_function, constraints=constraints, **kwargs)
        self.integer_vars = []
        self.binary_vars = []
        if integer_vars is not None:
            if not isinstance(integer_vars, list):
                integer_vars = [integer_vars]
            if any(isinstance(el, list) for el in integer_vars):
                raise TypeError("integer_vars list must not contain sub-lists")
            for int_idx in integer_vars:
                if not isinstance(int_idx, int) or int_idx < 0:
                    raise TypeError("integer_vars must be a list of non-negative indexes")
                self.integer_vars.append(int_idx)
        if binary_vars is not None:
            if not isinstance(binary_vars, list):
                binary_vars = [binary_vars]
            if any(isinstance(el, list) for el in binary_vars):
                raise TypeError("binary_vars must not contain sub-lists")
            for bin_idx in binary_vars:
                if not isinstance(bin_idx, int) or bin_idx < 0:
                    raise TypeError("binary_vars must be a list of non-negative indexes")
                self.binary_vars.append(bin_idx)

    def __get_cvxopt_objective_parameters(self):
        """Return the objective parameter required by cvxopt

        Returns:
            numpy.ndarray
        """
        c,  _ = self.objective_function.get_parameters()
        return c

    def __get_gurobi_objective_parameters(self):
        """Return the objective parameter required by cvxopt

        Returns:
            numpy.ndarray
        """
        c,  _ = self.objective_function.get_parameters()
        return c.flatten()

    def __get_cvxopt_constraints_parameters(self):
        """Return the constraints parameters required by cvxopt

        Returns:
            tuple: (A, b, G, h)
        """
        A = None
        b = None
        G = None
        h = None

        if len(self.aggregated_constraints) != 0:
            for const in self.aggregated_constraints:
                if const.is_equality:
                    A, b = const.get_parameters()
                    A = A.transpose()
                    b = -b
                    A = matrix(A, tc='d')
                    b = matrix(b, tc='d')

                if const.is_inequality:
                    G, h = const.get_parameters()
                    G = G.transpose()
                    h = -h
                    G = matrix(G, tc='d')
                    h = matrix(h, tc='d')

        return G, h, A, b

    def __get_gurobi_constraints_parameters(self):
        """Return the constraints parameters required by cvxopt

        Returns:
            tuple: (A, b, G, h)
        """
        A = None
        b = None
        G = None
        h = None

        if len(self.aggregated_constraints) != 0:
            for const in self.aggregated_constraints:
                if const.is_equality:
                    A, b = const.get_parameters()
                    A = A.transpose()
                    b = -b

                if const.is_inequality:
                    G, h = const.get_parameters()
                    G = G.transpose()
                    h = -h

        return G, h, A, b

    def __solve_cvxopt(self, solver='glpk', return_only_solution: bool = True):
        """Solve the problem through cvxopt interface

        Args:
            return_only_solution (bool): if True, returns only the optimal variable. If False, returns also the optimization status

        Raises:
            ValueError: solver must be 'glpk'
            ValueError: optimal solution not found

        Returns:
            numpy.ndarray: solution
        """
        try:
            import cvxopt
        except ImportError:
            raise ImportError("CVXOPT is required")
        try:
            import cvxopt.glpk
            options = dict(msg_lev='GLP_MSG_OFF')
        except ImportError:
            print("GLPK solver not found.")

        c = self.__get_cvxopt_objective_parameters()
        G, h, A, b = self.__get_cvxopt_constraints_parameters()

        sol = ilp(c, G, h, A, b, set(self.integer_vars), set(self.binary_vars), options)

        if sol[0] != 'optimal':
            raise ValueError("MILP optimum not found: %s" % sol[0])

        if not return_only_solution:
            if sol[0] == 'optimal':
                status = 'solved'
            else:
                status = 'unknown'

            output_dict = {
                'solution': np.array(sol[1]).reshape(self.input_shape),
                'status': status,
            }
            return output_dict
        else:
            return np.array(sol[1]).reshape(self.input_shape)

    def __solve_gurobi(self, initial_value: np.ndarray = None, return_only_solution: bool = True):
        """Solve the problem through gurobi

        Args:
            initial_value (np.ndarray, optional): starting point for warm start (if possible). Defaults to None.
            return_only_solution (bool): if True, returns only the optimal variable. If False, returns also the optimization status

        Raises:
            ImportError: Gurobi is required
            ValueError: optimal solution not found

        Returns:
            numpy.ndarray: solution
        """
        try:
            import gurobipy as gb
            from gurobipy import GRB
        except ImportError:
            raise ImportError("Gurobi is required")

        # create the model and variables
        model = gb.Model("milp")
        x = model.addMVar(self.input_shape[0], lb=-GRB.INFINITY)
        # set integer and binary constraints
        if self.integer_vars is not None:
            for ii in self.integer_vars:
                x[ii].setAttr(GRB.Attr.VType, GRB.INTEGER)
        if self.binary_vars is not None:
            for ii in self.binary_vars:
                x[ii].setAttr(GRB.Attr.VType, GRB.BINARY)
        if initial_value is not None:
            if not isinstance(initial_value, np.ndarray):
                raise TypeError("initial_value must be an instance of numpy.ndarray")
            if initial_value.ndim > 2 or (initial_value.ndim == 2 and initial_value.shape[1] > 1):
                raise ValueError("initial_value must be 1-dimensional or a 2-dimensional column vector")
            if initial_value.shape[0] != self.input_shape[0]:
                raise ValueError("The number of entries of initial_value and of the optimization variable must coincide")
            x[ii].setAttr("Start", initial_value[ii])

        # set objective function
        c = self.__get_gurobi_objective_parameters()
        model.setMObjective(None, c, 0.0, None, None, x, GRB.MINIMIZE)
        # set constraints and update the model
        G, h, A, b = self.__get_gurobi_constraints_parameters()
        if A is not None:
            model.addMConstrs(A, x, '=', b)
        if G is not None:
            model.addMConstrs(G, x, '<', h)

        model.setParam('OutputFlag', 0)
        model.update()

        model.optimize()

        sol = model.getAttr(GRB.Attr.Status)
        x_star = model.getAttr(GRB.Attr.X)
        if sol != 2:
            raise ValueError("MILP optimum not found: %d" % sol)
        if not return_only_solution:
            if sol == 2:
                status = 'solved'
            else:
                status = 'unknown'

            output_dict = {
                'solution': np.array(x_star).reshape(self.input_shape),
                'status': status,
            }
            return output_dict
        else:
            return np.array(x_star).reshape(self.input_shape)

    def solve(self, initial_value: np.ndarray = None, solver='glpk', return_only_solution: bool = True):
        """Solve the problem

        Args:
            initial_value (numpy.ndarray), optional): Initial value for warm start. Defaults to None. Not available in GLPK
            solver (str, optional): Solver to use ['glpk', 'gurobi']. Defaults to 'glpk'.

        Raises:
            ValueError: Unsupported solver

        Returns:
            numpy.ndarray: solution
        """

        if solver in ['glpk']:
            return self.__solve_cvxopt(return_only_solution=return_only_solution)
        elif solver in ['gurobi']:
            return self.__solve_gurobi(initial_value=initial_value, return_only_solution=return_only_solution)
        else:
            raise ValueError("Unsupported solver")



import numpy as np
import autograd.numpy as anp
import warnings
from .abstract_function import AbstractFunction
from .utilities import check_input


class Log(AbstractFunction):
    """Natural log function (elementwise)

    .. math::

        f(x)=\\log(x)

    with :math:`x: \\mathbb{R}^{n}`.

    Args:
        fn (AbstractFunction): input function

    Raises:
        TypeError: input must be a AbstractFunction object
    """

    def __init__(self, fn: AbstractFunction):
        if not isinstance(fn, AbstractFunction):
            raise TypeError("Input must be a AbstractFunction object")

        if not fn.is_differentiable:
            warnings.warn(
                'Composition with a nondifferentiable function will lead to an\
                    error when asking for a subgradient')
        else:
            self.differentiable = True

        self.fn = fn

        self.input_shape = fn.input_shape
        self.output_shape = fn.output_shape
        self.affine = False
        self.quadratic = False
        super().__init__()

    def _expression(self):
        expression = 'Log({})'.format(self.fn._expression())
        return expression 

    def _to_cvxpy(self):
        import cvxpy as cvx
        return cvx.log(self.fn._to_cvxpy())
    
    def _extend_variable(self, n_var, axis, pos):
        return Log(self.fn._extend_variable(n_var, axis, pos))

    @check_input
    def eval(self, x: np.ndarray) -> np.ndarray:
        return anp.log(self.fn.eval(x)).reshape(self.output_shape)

    @check_input
    def _alternative_jacobian(self, x: np.ndarray, **kwargs) -> np.ndarray:
        warnings.simplefilter("error")
        if not self.fn.is_differentiable:
            warnings.warn("Composition of non affine functions. The Jacobian may be not correct.")

        # (1)/(fn(x)) @ \Jac fn(x) 
        val = self.fn.eval(x).flatten()
        
        try:
            p1 = np.diag(1/val)
        except RuntimeWarning:
            # check for overflows
            p1 = np.diag(val)
            for idx, value in enumerate(val):
                if value > 600:
                    p1[idx, idx] = 0
                else:
                    p1[idx, idx] = 1/(value)

        p2 = self.fn.jacobian(x, **kwargs)
        return p1 @ p2



import numpy as np
from ..problems.linear_problem import LinearProblem
from ..functions import Variable

def generate_LP(n_var: int, n_constr: int, radius: float, direction: str='min', constr_form: str='ineq'):
    """Generate a feasible and not unbounded Linear Program and return problem data (cost, constraints, solution)
    TODO add reference
    
    Args:
        n_var (int): number of optimization variables
        n_constr (int): number of constraints
        radius (float): size of feasible set (for constr_form = 'ineq'), size of dual feasible set (for constr_form = 'eq')
        direction (str, optional): optimization direction - either 'max' (for maximization) or 'min' (for minimization). Defaults to 'min'
        constr_form (str, optional): form of constraints - either 'ineq' (for inequality: :math:`Ax \le b`) or 'eq' (for standard form: :math:`Ax = b, x \ge 0`). Defaults to 'ineq'

    Returns:
        tuple:
            - **c** (*numpy.ndarray*): cost vector
            - **A** (*numpy.ndarray*): constraint matrix
            - **b** (*numpy.ndarray*): constraint vector (right-hand side)
            - **solution** (*numpy.ndarray*): optimal solution of problem
    """

    size_1 = n_constr
    size_2 = n_var
    x = Variable(n_var)

    # swap dimensions if in dual form (equality constraints)
    if constr_form == 'eq':
        size_1, size_2 = size_2, size_1

    count = 0

    while True:
        count += 1

        # generate a problem in the form min_y c'y s.t. Ay <= b
        A_gen = np.random.randn(size_1, size_2)
        b_gen = np.random.randn(size_1, 1)
        c_gen = A_gen.transpose() @ np.random.randn(size_1, 1)

        # prepare problem data
        if constr_form == 'ineq':
            # primal problem
            A = A_gen
            b = b_gen
            c = c_gen
            constr = A_gen.transpose() @ x <= b_gen
        else:
            # dual problem: min_x b'x s.t. A'x = -c, x >= 0
            A = A_gen.transpose()
            b = -c_gen
            c = b_gen
            constr = [A_gen @ x == -c_gen, x >= 0]

        # form optimization problem object
        if direction == 'min':
            obj = c @ x
        else:
            obj = -c @ x

        prob = LinearProblem(objective_function=obj, constraints=constr)

        # solve problem to check whether an optimal solution exists
        try:
            solution = prob.solve()
            break
        except:
            continue

    return c, A, b, solution



import numpy as np
from autograd import grad as ag_grad
from autograd import jacobian as ag_jacobian
from autograd import hessian as ag_hessian
from typing import Union
from .utilities import check_input
import warnings


class AbstractFunction:
    """AbstractFunction class. This should be the parent of all specific (objective) functions.

    Attributes:
        input_shape (tuple): shape of the input of the function
        output_shape (tuple): shape of the output of the function
        differentiable (bool): True if the function is differentiable
        affine (bool): True if the function is affine
        quadratic (bool): True if the function is quadratic
    """
    # For forcing __r methods to be used when a numpy array comes first
    __array_priority__ = 1

    input_shape = None
    output_shape = None
    differentiable = False
    affine = False
    quadratic = False

    def __init__(self):
        self._setdiff()

    def _expression(self):
        return self.__class__.__name__

    def _to_cvxpy(self):
        pass

    def _extend_variable(self, n_var, axis, pos):
        pass

    def __str__(self):
        description = {'expression': self._expression(),
                       'input shape': self.input_shape,
                       'output shape': self.output_shape,
                       'affine': self.is_affine,
                       'quadratic': self.is_quadratic
                       }
        return str(description)

    def __neg__(self):
        return NegFunction(self)

    def __add__(self, other):
        if isinstance(other, AbstractFunction):
            return SumFunction(self, other)
        elif isinstance(other, (int, float, np.ndarray)):
            return ConstantSumFunction(other, self)
        else:
            raise TypeError

    def __radd__(self, other: Union[int, float, np.ndarray]):
        if isinstance(other, (int, float, np.ndarray)):
            return ConstantSumFunction(other, self)
        else:
            raise TypeError

    def __iadd__(self, other):
        if isinstance(other, AbstractFunction):
            return SumFunction(self, other)
        elif isinstance(other, (int, float, np.ndarray)):
            return ConstantSumFunction(other, self)
        else:
            raise TypeError

    def __sub__(self, other):
        if isinstance(other, AbstractFunction):
            return SumFunction(self, -other)
        elif isinstance(other, (int, float, np.ndarray)):
            return ConstantSumFunction(-other, self)
        else:
            raise TypeError

    def __rsub__(self, other):
        if isinstance(other, AbstractFunction):
            return SumFunction(-self, other)
        elif isinstance(other, (int, float, np.ndarray)):
            return ConstantSumFunction(other, -self)
        else:
            raise TypeError

    def __isub__(self, other):
        if isinstance(other, AbstractFunction):
            return SumFunction(self, -other)
        elif isinstance(other, (int, float, np.ndarray)):
            return ConstantSumFunction(-other, self)
        else:
            raise TypeError

    def __mul__(self, other):
        if isinstance(other, AbstractFunction):
            return MulFunction(self, other)
        elif isinstance(other, (int, float)):
            return ScalarMulFunction(other, self)
        else:
            raise TypeError

    def __rmul__(self, other: Union[int, float]):
        if isinstance(other, (int, float)):
            return ScalarMulFunction(other, self)
        else:
            raise NotImplementedError

    def __imul__(self, other):
        if isinstance(other, AbstractFunction):
            return MulFunction(self, other)
        elif isinstance(other, (int, float)):
            return ScalarMulFunction(other, self)
        else:
            raise TypeError
   
    def __truediv__(self, other):
        if isinstance(other, AbstractFunction):
            from .power import Power
            return MulFunction(self, Power(other, -1))
        elif isinstance(other, (int, float)):
            return ScalarMulFunction(1.0/other, self)
        else:
            raise TypeError

    def __rdiv__(self, other: Union[int, float]):
        if isinstance(other, (int, float)):
            return ScalarMulFunction(1.0/other, self)
        else:
            raise NotImplementedError

    def __idiv__(self, other):
        if isinstance(other, AbstractFunction):
            from .power import Power
            return MulFunction(self, Power(other, -1))
        elif isinstance(other, (int, float)):
            return ScalarMulFunction(1.0/other, self)
        else:
            raise TypeError

    def __pow__(self, other):
        if isinstance(other, (int, float)):
            from .power import Power
            return Power(self, float(other))
        else:
            raise TypeError

    def __ipow__(self, other):
        if isinstance(other, (int, float)):
            from .power import Power
            return Power(self, float(other))
        else:
            raise TypeError

    def __matmul__(self, other):
        if isinstance(other, AbstractFunction):
            return MatMulFunction(self, other)
        elif isinstance(other, (np.ndarray)):
            if other.shape == self.output_shape:
                return ConstantMatMulFunction(other, self)
            else:
                raise NotImplementedError
        else:
            raise TypeError

    def __rmatmul__(self, other: np.ndarray):
        if isinstance(other, (np.ndarray)):
            return ConstantMatMulFunction(other, self)
        else:
            raise TypeError

    def __imatmul__(self, other: np.ndarray):
        return self.__matmul__(other)

    def __to_constraint(self, other, sign):
        if not isinstance(other, (AbstractFunction, np.ndarray, float, int)):
            raise TypeError

        if isinstance(other, AbstractFunction):
            if (self.input_shape != other.input_shape) or (self.output_shape != other.output_shape):
                raise ValueError("Incompatible shapes")
            from ..constraints.constraints import Constraint
            return Constraint(self - other, sign=sign)

        if isinstance(other, np.ndarray):
            if other.shape != self.output_shape:
                raise ValueError("Incompatible shapes")
            from ..constraints.constraints import Constraint
            return Constraint(self - other, sign=sign)

        if isinstance(other, (int, float)):
            if self.output_shape != (1, 1):
                const = other * np.ones(self.output_shape)
            else:
                const = other
            from ..constraints.constraints import Constraint
            return Constraint(self - const, sign=sign)

    def __eq__(self, other):
        return self.__to_constraint(other, "==")

    def __le__(self, other):
        return self.__to_constraint(other, "<=")

    def __ge__(self, other):
        return self.__to_constraint(other, ">=")

    @property
    def is_differentiable(self):
        return self.differentiable

    @property
    def is_affine(self):
        return self.affine

    @property
    def is_quadratic(self):
        return self.quadratic

    def get_parameters(self):
        raise NotImplementedError

    def _setdiff(self):
        self._subgradient = ag_grad(self.eval)
        self._jacobian = ag_jacobian(self.eval)
        self._hessian = ag_hessian(self.eval)

    @check_input
    def eval(self, x: np.ndarray) -> np.ndarray:
        """Evaluate the function at a point x

        Args:
            x: input point
        """
        pass

    @check_input
    def jacobian(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """Evaluate the jacobian of the the function at a point x

        Args:
            x: input point
        """
        warnings.simplefilter("error", category=RuntimeWarning)
        warnings.simplefilter("ignore", category=UserWarning)  # TODO: check
        try:
            jac = np.squeeze(self._jacobian(x))
        except RuntimeWarning:
            try:
                jac = self._alternative_jacobian(x)
            except NotImplementedError:
                raise NotImplementedError("No jacobian can be computed")

        if jac.shape == ():  # in case of scalar
            jac = jac.reshape(1, 1)
        if len(jac.shape) == 1:
            jac = jac.reshape(1, -1)
        # np.nan_to_num(jac, copy=False)
        return jac

    @check_input
    def _alternative_jacobian(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """Evaluate the jacobian of the the function at a point x

        Args:
            x: input point
        """
        raise NotImplementedError

    @check_input
    def hessian(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """Evaluate the hessian of the the function at a point x

        Args:
            x: input point
        """
        warnings.simplefilter("error", category=RuntimeWarning)
        warnings.simplefilter("ignore", category=UserWarning)  # TODO: check
        try:
            hess = np.squeeze(self._hessian(x))
        except RuntimeWarning:
            try:
                hess = np.squeeze(self._alternative_hessian(x))
            except NotImplementedError:
                # TODO risolvere caso in cui autograd non riesce: np.true_divide(inf, inf)
                raise NotImplementedError("No hessian can be computed")

        if hess.shape == ():  # in case of scalar
            hess = hess.reshape(1, 1)
        # np.nan_to_num(hess, copy=False)
        return hess

    @check_input
    def _alternative_hessian(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """Evaluate the hessian of the the function at a point x

        Args:
            x: input point
        """
        raise NotImplementedError

    @check_input
    def subgradient(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """Evaluate the subgradient of the function at a point x

        Args:
            x: input point

        Raises:
            ValueError: subgradient is defined only for functions with scalar output 
        """
        warnings.simplefilter("error", category=RuntimeWarning)
        warnings.simplefilter("ignore", category=UserWarning)  # TODO: check
        if not self.output_shape == (1, 1):
            raise ValueError("Undefined subgradient")
        else:
            try:
                subg = self._alternative_subgradient(x).reshape(self.input_shape)
            except NotImplementedError:
                try:
                    subg = self._subgradient(x).reshape(self.input_shape)
                    # subg = self._alternative_subgradient(x).reshape(self.input_shape)
                except RuntimeWarning:
                    raise NotImplementedError("No subgradient can be computed")
            # subg = self.jacobian(x).reshape(self.input_shape)
            # np.nan_to_num(subg, copy=False)
            return subg

    @check_input
    def _alternative_subgradient(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """Evaluate the hessian of the the function at a point x

        Args:
            x: input point
        """
        return self._alternative_jacobian(x).reshape(self.input_shape)


class ConstantSumFunction(AbstractFunction):
    """Add a constant to a AbstractFunction

    Args:
        constant (numpy.ndarray): constant to add
        fn (AbstractFunction): function to add a constant to

    Raises:
        ValueError: The shape of the constant must be equal to the output shape of the function
    """

    def __init__(self, constant: Union[int, float, np.ndarray], fn: AbstractFunction):
        self.constant = constant
        if isinstance(constant, np.ndarray):
            if fn.output_shape != constant.shape:
                raise ValueError("Different shapes")
        self.fn = fn
        self.input_shape = fn.input_shape
        self.output_shape = fn.output_shape
        self.differentiable = fn.is_differentiable
        if fn.is_affine:
            self.affine = True
        if fn.is_quadratic:
            self.quadratic = True

        super().__init__()

    def _expression(self):
        return str(self.constant) + " + " + self.fn._expression()

    @check_input
    def eval(self, x):
        return self.fn.eval(x) + self.constant

    def _to_cvxpy(self):
        return self.constant + self.fn._to_cvxpy()
    
    def _extend_variable(self, n_var, axis, pos):
        return self.fn._extend_variable(n_var, axis, pos) + self.constant

    @check_input
    def jacobian(self, x):
        return self.fn.jacobian(x)

    @check_input
    def hessian(self, x):
        return self.fn.hessian(x)

    @check_input
    def subgradient(self, x):
        return self.fn.subgradient(x)


class SumFunction(AbstractFunction):
    """Sum two AbstractFunction objects

    Args:
        f1 (AbstractFunction): first function
        f2 (AbstractFunction): second function

    Raises:
        ValueError: The input/output shapes of the two functions must be the same
    """

    def __init__(self, f1: AbstractFunction, f2: AbstractFunction):
        if f1.input_shape != f2.input_shape or f1.output_shape != f2.output_shape:
            raise ValueError("Different input/output shapes")
        self.f1 = f1
        self.f2 = f2
        self.input_shape = f1.input_shape
        self.output_shape = f1.output_shape
        if f1.is_differentiable and f2.is_differentiable:
            self.differentiable = True
        if f1.is_affine and f2.is_affine:
            self.affine = True
        if (f1.is_quadratic and f2.is_quadratic) or (f1.is_quadratic and f2.is_affine) or (f1.is_affine and f2.is_quadratic):
            self.quadratic = True

        super().__init__()

    def _expression(self):
        return self.f1._expression() + " + " + self.f2._expression()

    def _to_cvxpy(self):
        return self.f1._to_cvxpy() + self.f2._to_cvxpy()
    
    def _extend_variable(self, n_var, axis, pos):
        return self.f1._extend_variable(n_var, axis, pos) + self.f2._extend_variable(n_var, axis, pos)

    @check_input
    def eval(self, x):
        return self.f1.eval(x) + self.f2.eval(x)

    @check_input
    def jacobian(self, x):
        return self.f1.jacobian(x) + self.f2.jacobian(x)

    @check_input
    def hessian(self, x):
        return self.f1.hessian(x) + self.f2.hessian(x)

    @check_input
    def subgradient(self, x):
        return self.f1.subgradient(x) + self.f2.subgradient(x)


class NegFunction(AbstractFunction):
    """Multiplies a AbstractFunction by -1

    Args:
        fn (AbstractFunction): input function function
    """

    def __init__(self, fn: AbstractFunction):
        self.fn = fn
        self.input_shape = fn.input_shape
        self.output_shape = fn.output_shape
        self.differentiable = fn.is_differentiable
        if fn.is_affine:
            self.affine = True

        if fn.is_quadratic:
            self.quadratic = True

        super().__init__()

    def _expression(self):
        return " - " + self.fn._expression()

    def _to_cvxpy(self):
        return -self.fn._to_cvxpy()
    
    def _extend_variable(self, n_var, axis, pos):
        return -self.fn._extend_variable(n_var, axis, pos)

    @check_input
    def eval(self, x):
        return -self.fn.eval(x)

    @check_input
    def jacobian(self, x):
        return -self.fn.jacobian(x)

    @check_input
    def hessian(self, x):
        return -self.fn.hessian(x)

    @check_input
    def subgradient(self, x):
        return -self.fn.subgradient(x)


class MulFunction(AbstractFunction):
    def __init__(self, f1: AbstractFunction, f2: AbstractFunction):
        """Multiplies two functions (with scalar outputs)

        Args:
            f1 (AbstractFunction): first function
            f2 (AbstractFunction): second function

        Raises:
            ValueError: input functions must have the same input shapes
            ValueError: at least the first input function must have scalar output
            NotImplementedError: only differentiable input functions are currently supported
        """
        if f1.input_shape != f2.input_shape:
            raise ValueError("Different input shapes")
        if not(f1.output_shape == (1, 1)):
            raise ValueError(
                "At least the first input function must have scalar output")
        self.f1 = f1
        self.f2 = f2
        self.input_shape = f1.input_shape
        self.output_shape = f2.output_shape
        if f1.is_differentiable and f2.is_differentiable:
            self.differentiable = True

        super().__init__()

    def _expression(self):
        return self.f1._expression() + " * " + self.f2._expression()

    @check_input
    def eval(self, x):
        return self.f1.eval(x) * self.f2.eval(x)

    def _extend_variable(self, n_var, axis, pos):
        return self.f1._extend_variable(n_var, axis, pos) * self.f2._extend_variable(n_var, axis, pos)

    def _to_cvxpy(self):
        return self.f1._to_cvxpy() * self.f2._to_cvxpy()
    

class ScalarMulFunction(AbstractFunction):
    def __init__(self, scalar: Union[int, float], fn: AbstractFunction):
        """Multiply a function by a scalar

        Args:
            scalar (float or int): scalar
            fn (AbstractFunction): input function
        Raises:
            ValueError: the scalar must be a float (or an int)
        """
        if not isinstance(scalar, (int, float)):
            raise ValueError(
                "Multiplication by scalars requires the scalar to be float or int")
        self.scalar = scalar
        self.fn = fn
        self.input_shape = fn.input_shape
        self.output_shape = fn.output_shape
        if fn.is_differentiable:
            self.differentiable = True
        if fn.is_affine:
            self.affine = True
        if fn.is_quadratic:
            self.quadratic = True

        super().__init__()

    def _expression(self):
        return str(self.scalar) + " * " + self.fn._expression()

    def _to_cvxpy(self):
        return self.scalar * self.fn._to_cvxpy()

    @check_input
    def eval(self, x):
        return self.scalar * self.fn.eval(x)
    
    def _extend_variable(self, n_var, axis, pos):
        return self.scalar * self.fn._extend_variable(n_var, axis, pos)

    @check_input
    def jacobian(self, x):
        return self.scalar * self.fn.jacobian(x)

    @check_input
    def hessian(self, x):
        return self.scalar * self.fn.hessian(x)

    @check_input
    def subgradient(self, x):
        return self.scalar * self.fn.subgradient(x)


class MatMulFunction(AbstractFunction):
    def __init__(self, f1: AbstractFunction, f2: AbstractFunction):
        """Dot product of two function

        :math:`f(x) = \\langle f_1(x), f_2(x)\\rangle=\\f_1(x)^\\top f_2(x)

        with :math:`f_1,f_2:\\mathbb{R}^m\\to\\mathbb{R}^n`

        Args:
            f1 (AbstractFunction): first function (output shape (n, 1))
            f2 (AbstractFunction): second function (output shape (n, 1))

        Raises:
            ValueError: input functions must have the same input shapes
            NotImplementedError: only differentiable input functions are currently supported
        """
        if not (f1.input_shape == f2.input_shape and f1.output_shape == f2.output_shape):
            raise ValueError("Different input/output shapes")

        self.f1 = f1
        self.f2 = f2
        self.input_shape = f1.input_shape
        self.output_shape = (1, 1)

        if f1.output_shape[1] != 1:
            raise ValueError("Unsupported input functions output shape. Must be (n, 1)")

        if f1.is_differentiable and f2.is_differentiable:
            self.differentiable = True

        super().__init__()

    def _expression(self):
        return self.f1._expression() + " @ " + self.f2._expression()

    @check_input
    def eval(self, x):
        return self.f1.eval(x).transpose() @ self.f2.eval(x)

    def _extend_variable(self, n_var, axis, pos):
        return MatMulFunction(self.f1._extend_variable(n_var, axis, pos), self.f2._extend_variable(n_var, axis, pos))

    def _to_cvxpy(self):
        return self.f1._to_cvxpy().transpose() @ self.f2._to_cvxpy()


class ConstantMatMulFunction(AbstractFunction):
    def __init__(self, constant: np.ndarray, fn: AbstractFunction):
        """Matrix/Dot product of a matrix/vector and a function

        :math:`f(x) = \\langle A, g(x)\\rangle=\\A^\\top g(x)

        with :math:`A:\\mathbb{R}^{n\\times m}` and :math:`g:\\mathbb{R}^d\\to\\mathbb{R}^n`

        Args:
            constant (numpy.ndarray): constant vector (shape (n, m))
            fn (AbstractFunction): function (output shape (n, 1))

        Raises:
            ValueError: input functions must have the same input shapes
            NotImplementedError: only differentiable input functions are currently supported
        """
        if not isinstance(constant, (np.ndarray)):
            raise ValueError(
                "Multiplication by constant vector requires the vector to be a numpy.ndarray")
        if fn.output_shape[0] != constant.shape[0]:
            raise ValueError(
                "Incompatible dimensions: function ouput is {} and given matrix is {}".format(
                    fn.output_shape, constant.shape))

        self.constant = constant
        self.fn = fn
        self.input_shape = fn.input_shape
        self.output_shape = (self.constant.shape[1], self.fn.output_shape[1])

        if fn.output_shape[1] != 1:
            raise ValueError("Unsupported function output shape. Must be (n, 1)")

        if fn.is_differentiable:
            self.differentiable = True

        if fn.is_affine:
            self.affine = True

        super().__init__()

    def _expression(self):
        return str(self.constant()) + " @ " + self.fn._expression()

    @check_input
    def eval(self, x):
        return self.constant.transpose() @ self.fn.eval(x)

    def _extend_variable(self, n_var, axis, pos):
        return ConstantMatMulFunction(self.constant, self.fn._extend_variable(n_var, axis, pos))

    def _to_cvxpy(self):
        return self.constant.transpose() @ self.fn._to_cvxpy()



import numpy as np
from copy import deepcopy
from ..agents.agent import Agent
from ..problems.problem import Problem
from .algorithm import Algorithm
from ..functions import Variable
from ..functions.squared_norm import SquaredNorm
from ..functions.quadratic_form import QuadraticForm


class ADMM(Algorithm):
    """Distributed ADMM.

    From the perspective of agent :math:`i` the algorithm works as follows.

    Initialization: :math:`\lambda_{ij}^0` for all :math:`j \in \mathcal{N}_i`, :math:`\lambda_{ii}^0` and :math:`z_i^0`

    For :math:`k=0,1,\\dots`

    * Compute :math:`x_i^k` as the optimal solution of

    .. math::
        \min_{x_i \in X_i} \: f_i(x_i) + x_i^\\top \sum_{j \in \mathcal{N}_i \cup \\{i\\}} \lambda_{ij}^k + \\frac{\\rho}{2} \sum_{j \in \mathcal{N}_i \cup \\{i\\}} \\| x_i - z_j^t \\|^2

    * Gather :math:`x_j^k` and :math:`\lambda_{ji}^k` from neighbors :math:`j \in \mathcal{N}_i`

    * Update :math:`z_i^{k+1}`

    .. math::
        z_i^{k+1} = \\frac{\sum_{j \in \mathcal{N}_i \cup \\{i\\}} x_j^k}{|\mathcal{N}_i| + 1} + \\frac{\sum_{j \in \mathcal{N}_i \cup \\{i\\}} \lambda_{ji}^k}{\\rho (|\mathcal{N}_i| + 1)}

    * Gather :math:`z_j^{k+1}` from neighbors :math:`j \in \mathcal{N}_i`

    * Update for all :math:`j \in \mathcal{N}_i`

    .. math::
        \lambda_{ij}^{k+1} = \lambda_{ij}^{k} + \\rho (x_i^k - z_j^{k+1})

    where :math:`x_i, z_i\\in\\mathbb{R}^{n}`, :math:`\lambda_{ij}\\in\\mathbb{R}^n` for all :math:`j \in \mathcal{N}_i \cup \\{i\\}`, :math:`X_i\\subseteq\mathbb{R}^{n}` for all :math:`i` and :math:`\\rho` is a positive penalty parameter.

    The algorithm has been presented in ????.
    """

    # TODO choose ref
    def __init__(self, agent: Agent, initial_lambda: dict, initial_z: np.ndarray, enable_log: bool = False):
        super(ADMM, self).__init__(agent, enable_log)

        if not isinstance(agent.problem, Problem):
            raise TypeError("The agent must be equipped with a Problem")

        if sum(1 for i in agent.problem.objective_function.input_shape if i > 1) > 1:
            raise ValueError("Currently only mono-dimensional objective functions are supported")

        if not all([isinstance(x, np.ndarray) for _, x in initial_lambda.items()]):
            raise TypeError("The initial condition dictionary can only contain numpy vectors")

        # augmented set of neighbors and degree
        self.augmented_neighbors = deepcopy(self.agent.in_neighbors)
        self.augmented_neighbors.append(agent.id)
        self.augmented_neighbors.sort()
        self.degree = len(agent.in_neighbors)

        if self.augmented_neighbors != sorted([x for x in initial_lambda]):
            raise TypeError(
                "The initial condition dictionary must contain exactly one vector per neighbor (plus the agent itself)")

        # shape of local variable
        self.x_shape = agent.problem.objective_function.input_shape

        # initialize dual variables and primal solution
        self.lambd0 = deepcopy(initial_lambda)
        self.lambd = deepcopy(initial_lambda)
        self.z0 = initial_z
        self.z = initial_z
        self.x = None

    def run(self, iterations: int = 1000, penalty: float = 0.1, verbose: bool = False, **kwargs) -> np.ndarray:
        """Run the algorithm for a given number of iterations

        Args:
            iterations: Number of iterations. Defaults to 1000.
            penalty: ADMM penalty parameter. Defaults to 0.1.
            verbose: If True print some information during the evolution of the algorithm. Defaults to False.

        Raises:
            TypeError: The number of iterations must be an int

        Returns:
            return a tuple (x, lambda, z) with the sequence of primal solutions, dual variables and auxiliary primal variables if enable_log=True.
        """
        if not isinstance(iterations, int):
            raise TypeError("The number of iterations must be an int")

        if self.enable_log:
            # initialize sequence of x and z
            x_dims = [iterations]
            for dim in self.x_shape:
                x_dims.append(dim)
            self.x_sequence = np.zeros(x_dims)
            self.z_sequence = np.zeros(x_dims)

            # initialize sequence of lambda
            self.lambda_sequence = {}
            for j in self.augmented_neighbors:
                self.lambda_sequence[j] = np.zeros(x_dims)

        self.initialize_algorithm()

        for k in range(iterations):
            # store current lambda and z
            if self.enable_log:
                for j in self.augmented_neighbors:
                    self.lambda_sequence[j][k] = self.lambd[j]
                self.z_sequence[k] = self.z

            self.iterate_run(rho=penalty, **kwargs)

            # store primal solution
            if self.enable_log:
                self.x_sequence[k] = self.x

            if verbose:
                if self.agent.id == 0:
                    print('Iteration {}'.format(k), end="\r")

        if self.enable_log:
            return (self.x_sequence, self.lambda_sequence, self.z_sequence)

    def _update_local_solution(self, x: np.ndarray, z: np.ndarray, z_neigh: dict, rho: float, **kwargs):
        """Update the local solution

        Args:
            x: current solution
            z: current auxiliary primal variable
            z_neigh: auxiliary primal variables of neighbors (dictionary)
            rho: penalty parameter
        """
        self.z_neigh = z_neigh

        # update dual variables
        self.lambd[self.agent.id] += rho * (x - z)
        for j, z_j in z_neigh.items():
            self.lambd[j] += rho * (x - z_j)

        # update primal variables
        self.x = x
        self.z = z

    def initialize_algorithm(self):
        """Initializes the algorithm
        """
        # exchange z with neighbors
        self.z_neigh = self.agent.neighbors_exchange(self.z)

    def iterate_run(self, rho: float, **kwargs):
        """Run a single iterate of the algorithm
        """
        # TODO extend to non mono-dimensional variables

        # build local problem
        x_i = Variable(self.x_shape[0])

        penalties = [(x_i - self.z_neigh[j])@(x_i - self.z_neigh[j]) for j in self.agent.in_neighbors]
        penalties.append((x_i - self.z)@(x_i - self.z))
        sumlambda = sum(self.lambd.values())

        obj_function = self.agent.problem.objective_function + sumlambda @ x_i + (rho/2) * sum(penalties)
        pb = Problem(obj_function, self.agent.problem.constraints)

        # solve problem and save data
        x = pb.solve()

        # exchange primal variables and dual variables with neighbors
        x_neigh = self.agent.neighbors_exchange(x)
        lambda_neigh = self.agent.neighbors_exchange(self.lambd, dict_neigh=True)

        # compute auxiliary variable
        z = (sum(x_neigh.values()) + x) / (self.degree+1) + \
            (sum(lambda_neigh.values()) + self.lambd[self.agent.id]) / (rho*(self.degree+1))

        # exchange auxiliary variables with neighbors
        z_neigh = self.agent.neighbors_exchange(z)

        # update local data
        self._update_local_solution(x, z, z_neigh, rho, **kwargs)

    def get_result(self):
        """Return the current value primal solution, dual variable and auxiliary primal variable

        Returns:
            tuple (primal, dual, auxiliary): value of primal solution (np.ndarray), dual variables (dictionary of np.ndarray), auxiliary primal variable (np.ndarray)
        """
        return (self.x, self.lambd, self.z)



import numpy as np
from .problem import Problem
from .quadratic_problem import QuadraticProblem


class ProjectionProblem(Problem):
    """Computes the projection of a point onto some constraints, i.e., it solves

    .. math::

        \\text{minimize } & \\frac{1}{2}\\| x - p \\|^2 

        \\text{subject to } & f_k(x)\\leq 0, \\, k=1,...


    Args:
        constraints_list (list): list of constraints
        point (numpy.ndarray): point :math:`p` to project
    """

    def __new__(cls, *args, **kwargs):
        instance = object.__new__(cls)
        return instance

    def __init__(self, constraints, point, **kwargs):
        from ..functions import Variable, QuadraticForm
        x = Variable(point.shape[0])
        objective_function = 0.5 * (x - point) @ (x - point)
        self.point = point

        super(ProjectionProblem, self).__init__(objective_function, constraints, **kwargs)

        self.only_affine_constraints = False
        if len(self.nonlinear_constraints) == 0:
            self.only_affine_constraints = True

    def solve(self):
        """solve the problem

        Returns:
            numpy.ndarray: solution
        """
        satisfied = True
        if self.affine_equality:
            satisfied &= self.affine_equality.eval(self.point)
        if self.affine_inequality:
            satisfied &= self.affine_inequality.eval(self.point)
        for constr in self.nonlinear_constraints:
            satisfied &= constr.eval(self.point)

        if satisfied:
            return self.point
        else:
            if self.only_affine_constraints:
                # if non fat matrix, projection is explicit
                A_eq = None
                A_ineq = None
                if self.affine_equality:
                    A_eq, b_eq = self.affine_equality.get_parameters()
                if self.affine_inequality:
                    A_ineq, b_ineq = self.affine_inequality.get_parameters()
                if (A_eq is not None) and (A_ineq is not None):
                    A = np.hstack([A_eq, A_ineq])
                    b = np.vstack([b_eq, b_ineq])
                elif A_eq is not None:
                    A = A_eq
                    b = b_eq
                else:
                    A = A_ineq
                    b = b_ineq
                if A.shape[0] >= A.shape[1]:
                    projected_pt = self.point - A @ np.linalg.inv(A.transpose() @ A) @ (A.transpose() @ self.point + b)
                    return projected_pt
                else:
                    from .quadratic_problem import QuadraticProblem
                    qp = QuadraticProblem(self.objective_function, self.constraints, is_pos_def=True)
                    return qp.solve()
            else:
                return super().solve()



import numpy as np
from . import MixedIntegerLinearProblem, LinearProblem
from ..functions import AffineForm, Variable
from ..functions.affine_form import aggregate_affine_form
from ..constraints import Constraint, AbstractSet


class ConvexifiedMILP(MixedIntegerLinearProblem):
    """Solve a convexified Mixed-Integer Linear Problem of the form:

    .. math::

        \\text{minimize } & c^\\top x + M \\rho

        \\text{subject to } & x \\in \mathrm{conv}(X), \:\\rho \\geq 0

                            & Ax \\leq y + \\rho \mathbf{1}
        
    where the set :math:`X` is a compact mixed-integer polyhedral set defined by equality
    and inequality constraints. Moreover, :math:`\\rho` is a scalar,
    :math:`y \in \mathbb{R}^{m}` is a vector and :math:`A \in \mathbb{R}^{m \\times n}`
    is a constraint matrix.
    """

    def __init__(self, objective_function: AffineForm, y: np.ndarray, A: np.ndarray,
        constraints: list = None, integer_vars: list = None, binary_vars: list = None, **kwargs):
        """initialize the class

        Args:
            objective_function (AffineForm): linear cost function :math:`c^\\top x`
            y (np.ndarray): value of the vector :math:`y`
            A (np.ndarray): constraint matrix
            constraints (list): equality and inequality constraints describing the set :math:`X`
            integer_vars (list, optional): list of integer variables in the set :math:`X`
            binary_vars (list, optional): list of binary variables in the set :math:`X`
        """

        super().__init__(objective_function=objective_function, constraints=constraints,
            integer_vars=integer_vars, binary_vars=binary_vars, **kwargs)

        # check inputs
        if not isinstance(y, np.ndarray):
            raise TypeError("y must be an instance of numpy.ndarray")
        if not isinstance(A, np.ndarray):
            raise TypeError("A must be an instance of numpy.ndarray")
        
        # check dimensions
        if y.ndim > 2 or (y.ndim == 2 and y.shape[1] > 1):
            raise ValueError("y must be 1-dimensional or a 2-dimensional column vector")
        if A.ndim != 2:
            raise ValueError("A must be 2-dimensional matrix")
        if y.shape[0] != A.shape[0]:
            raise ValueError("The number of rows of y and A must coincide")
        if objective_function.input_shape[0] != A.shape[1]:
            raise ValueError("The number of variables in objective_function must coincide with the number of columns of A")

        # save data
        self.nvars   = A.shape[1] # number of optimization variables
        self.nconstr = A.shape[0] # number of inequality constraints
        self.y = np.copy(y)
        self.A = np.copy(A)
        if y.ndim == 1:
            self.y = self.y[:, None]

        # initialize more data
        self.last_mu = np.zeros((self.nconstr, 1))
        self.x = Variable(self.nvars) # symbolic primal variable
        self.dual_var = Variable(self.nconstr + 1) # symbolic dual variable: [mu, epigraph]

        # prepare MILP problem
        self.c = objective_function.get_parameters()[0] # primal cost vector
        self.lagr_cost = (self.c + self.A.T @ self.last_mu) @ self.x # cost function of Lagrangian
        self.milp = MixedIntegerLinearProblem(objective_function=self.lagr_cost, constraints=constraints,
            integer_vars=integer_vars, binary_vars=binary_vars) # lagrangian MILP problem
        
        # prepare LP problem
        self.lp_cost = np.vstack((self.y, 1)) @ self.dual_var
    
    def solve(self, M: float, milp_solver=None, initial_dual_solution: np.ndarray = None,
        return_only_solution: bool = True, y: np.ndarray = None, max_cutting_planes: int = None,
        cutting_planes: np.ndarray = None, threshold_convergence: float = 1e-8, max_iterations: int = 1000):
        """Solve the problem using a custom dual cutting-plane algorithm

        Args:
            M (float): value of the parameter :math:`M`
            milp_solver (str, optional): MILP solver to use. Defaults to None (use default solver).
            initial_dual_solution (numpy.ndarray, optional): Initial dual value for warm start. Defaults to None.
            return_only_solution (bool, optional): if True, returns only solution, otherwise returns more information. Defaults to True.
            y (np.ndarray, optional): value to override the current y. Defaults to None (keep current value).
            max_cutting_planes (int, optional): maximum number of stored cutting planes. Defaults to None (disabled).
            cutting_planes (numpy.ndarray, optional): cutting planes for warm start previously returned by this function. Defaults to None.
            threshold_convergence (float, optional): threshold for convergence detection. Defaults to 1e-8.
            max_iterations (int, optional): maximum number of iterations performed by algorithm. Defaults to 1e3.

        Returns:
            tuple: primal solution tuple (x, rho) if return_only_solution = True,
            otherwise (primal solution tuple, optimal cost, dual solution, cutting planes, number of iterations)
        """

        # check parameters
        if max_cutting_planes is not None and (not isinstance(max_cutting_planes, int) or max_cutting_planes <= 0):
            raise TypeError("max_cutting_planes must be either None or a positive integer")
        if not isinstance(max_iterations, int) or max_iterations <= 0:
            raise TypeError("max_iterations must be a positive integer")
        if not isinstance(threshold_convergence, float) or threshold_convergence <= 0:
            raise TypeError("threshold_convergence must be a positive float")
        if milp_solver is not None and not isinstance(milp_solver, str):
            raise TypeError("milp_solver must be either None or a string")
        if not isinstance(M, float) or M < 0:
            raise TypeError("M must be a non-negative float")

        # check initial dual solution
        if initial_dual_solution is not None:
            if not isinstance(initial_dual_solution, np.ndarray):
                raise TypeError("initial_dual_solution must be either None or a numpy.ndarray")
            if initial_dual_solution.ndim > 2 or (initial_dual_solution.ndim == 2 and initial_dual_solution.shape[1] > 1):
                raise ValueError("initial_dual_solution must be 1-dimensional or a 2-dimensional column vector")
            if initial_dual_solution.shape[0] != self.nconstr:
                raise ValueError("The number of rows of initial_dual_solution and A must coincide")

            if initial_dual_solution.ndim == 1:
                initial_dual_solution = np.copy(initial_dual_solution)[:, None]
            else:
                initial_dual_solution = np.copy(initial_dual_solution)

        # check y
        if y is not None:
            if not isinstance(y, np.ndarray):
                raise TypeError("y must be an instance of numpy.ndarray")
            if y.ndim > 2 or (y.ndim == 2 and y.shape[1] > 1):
                raise ValueError("y must be 1-dimensional or a 2-dimensional column vector")
            if y.shape[0] != self.nconstr:
                raise ValueError("The number of rows of y and A must coincide")

            if y.ndim == 1:
                self.y = np.copy(y)[:, None]
            else:
                self.y = np.copy(y)
        
        # check cutting planes
        # cutting planes are row vectors of the form [A, b, x^T]
        # (each one represents a constraint Ax <= b obtained at MILP solution x)
        n_cut = 0
        if cutting_planes is not None:
            if not isinstance(cutting_planes, np.ndarray):
                raise TypeError("cutting_planes must be either None or a numpy.ndarray")
            if cutting_planes.ndim != 2:
                raise ValueError("cutting_planes must be a 2-dimensional matrix")
            if cutting_planes.shape[1] != self.nconstr + self.nvars + 2:
                raise ValueError("The number of columns of cutting_planes must coincide with the sum of sizes of A + 2")

            n_cut = cutting_planes.shape[0]

        # initialize dual variable
        if initial_dual_solution is not None:
            # project onto positive orthant
            mu = np.maximum(initial_dual_solution, np.zeros((self.nconstr, 1)))
        else:
            mu = self.last_mu
        
        # check maximum number of allowed cutting planes
        n_allowed_cuts = max_iterations 
        if max_cutting_planes is not None:
            n_allowed_cuts = min(max_iterations, max_cutting_planes)

            # discard excess cutting planes if necessary
            n_cut = min(n_cut, n_allowed_cuts-1) # pro tip: -1 to leave one empty slot for initialization

        # initialize storage variables
        x_store = np.zeros((self.nvars, n_allowed_cuts))
        A_cuts  = np.zeros((n_allowed_cuts, self.nconstr + 1)) # cutting planes
        b_cuts  = np.zeros((n_allowed_cuts, 1))

        # constraints mu^T 1 <= M and mu >= 0
        self.M = M
        basic_lp_constraints = [
                np.column_stack((np.ones((1, self.nconstr)), 0)).T @ self.dual_var <= self.M,
                np.column_stack((-np.eye(self.nconstr), np.zeros((self.nconstr, 1)))).T @ self.dual_var <= 0
        ]
        
        # initialize cutting planes
        if cutting_planes is not None:
            # take only the last n_cut cuts
            A_cuts[0:n_cut, :]  = cutting_planes[-n_cut:,0:self.nconstr+1]
            b_cuts[0:n_cut]     = cutting_planes[-n_cut:,self.nconstr+1:self.nconstr+2]
            x_store[:, 0:n_cut] = cutting_planes[-n_cut:,-self.nvars:].T
        
        # initialize LP problem
        self.lp_cost = np.row_stack((self.y, 1)) @ self.dual_var
        
        ###########################################################

        # solve initial MILP
        self.lagr_cost = (self.c + self.A.T @ mu) @ self.x
        self.milp.objective_function = self.lagr_cost
        x = self.milp.solve() if milp_solver is None else self.milp.solve(solver=milp_solver)
        
        x_store[:, n_cut] = x.flatten()
        A_cuts[n_cut, :]  = np.column_stack((-(self.A @ x).T, -1))
        b_cuts[n_cut]     = self.c.T @ x
        true_cost         = self.lagr_cost.eval(x) - mu.T@self.y

        # perform iterations
        for k in range(max_iterations):
            # compute last index of cutting planes that can be used
            index_k = min(n_cut+k+1, n_allowed_cuts)

            # solve LP
            lp_constr = [A_cuts[0:index_k, :].T @ self.dual_var <= b_cuts[0:index_k]]
            lp_constr.extend(basic_lp_constraints)
            lp = LinearProblem(objective_function=self.lp_cost, constraints=lp_constr)
            solution = lp.solve(return_only_solution=False)
            mu = solution['solution'][:-1] # mu variable
            multipliers = solution['dual_variables']
            approx_cost = -self.lp_cost.eval(solution['solution'])

            # solve MILP
            self.lagr_cost = (self.c + self.A.T @ mu) @ self.x
            self.milp.objective_function = self.lagr_cost
            x = self.milp.solve() if milp_solver is None else self.milp.solve(solver=milp_solver)
            
            true_cost = self.lagr_cost.eval(x) - mu.T@self.y
            
            # check stopping criterion
            if (abs(true_cost - approx_cost) < threshold_convergence):
                break
            
            # in case of saturation, free a slot for next cutting plane
            if index_k >= n_allowed_cuts:
                A_cuts  = np.roll(A_cuts, -1, axis=0)
                b_cuts  = np.roll(b_cuts, -1, axis=0)
                x_store = np.roll(x_store, -1, axis=1)
                index_k -= 1
            
            # store new cutting plane
            x_store[:, index_k] = x.flatten()
            A_cuts[index_k, :]  = np.column_stack((-(self.A @ x).T, -1))
            b_cuts[index_k]     = self.c.T @ x
        
        # save data and compute primal solution
        self.last_mu = mu
        lambdas  = multipliers[0].flatten() # multipliers associated with cutting planes
        rho_star = np.asscalar(multipliers[1].flatten()) # multiplier associated with mu^T 1 <= M
        x_star   = sum([lambdas[i]*x_store[:, i] for i in range(index_k)]) # primal solution
        x_star   = x_star[:, None]

        # prepare cutting planes to return (duplicates are removed)
        cutting_planes = np.column_stack((A_cuts[0:index_k, :], b_cuts[0:index_k, :], x_store[:, 0:index_k].T))
        unique_ind     = np.unique(cutting_planes, axis=0, return_index=True)[1]
        cutting_planes = cutting_planes[unique_ind, :]

        # return solution
        if not return_only_solution:
            # TODO unify this return with the returns of other files
            # return a tuple with (primal solution, optimal cost, dual solution, cutting planes, number of iterations)
            return ((x_star, rho_star), np.asscalar(true_cost), mu, cutting_planes, k+1)
        else:
            # return only primal solution (x, rho)
            return (x_star, rho_star)


import numpy as np
import autograd.numpy as anp
import warnings
from .abstract_function import AbstractFunction
from .utilities import check_input


class Square(AbstractFunction):
    """Square function (elementwise)

    .. math::

        f(x)= x^2

    with :math:`x: \\mathbb{R}^{n}`.

    Args:
        fn (AbstractFunction): input function

    Raises:
        TypeError: input must be a AbstractFunction object
    """

    def __init__(self, fn: AbstractFunction):
        if not isinstance(fn, AbstractFunction):
            raise TypeError("Input must be a AbstractFunction object")

        if fn.is_differentiable:
            self.differentiable = True

        self.fn = fn

        self.input_shape = fn.input_shape
        self.output_shape = fn.output_shape

        self.affine = False
        self.quadratic = False # TODO

        super().__init__()
        
    def _expression(self):
        expression = 'Square({})'.format(self.fn._expression())
        return expression 

    def _to_cvxpy(self):
        import cvxpy as cvx
        return cvx.square(self.fn._to_cvxpy())

    @check_input
    def eval(self, x: np.ndarray) -> np.ndarray:
        return anp.power(self.fn.eval(x), 2).reshape(self.output_shape)
    
    def _extend_variable(self, n_var, axis, pos):
        return Square(self.fn._extend_variable(n_var, axis, pos))

    # @check_input
    # def jacobian(self, x: np.ndarray, **kwargs) -> np.ndarray:
    #     if not self.fn.is_differentiable:
    #         warnings.warn("Composition of non affine functions. The Jacobian may be not correct.")
    #     # 2 fn(x) \jac fn(x)
    #     val = self.fn.eval(x)
    #     p1 = 2 * np.diag(val.flatten())
    #     p2 = self.fn.jacobian(x, **kwargs)
    #     return p1 @ p2



import numpy as np
from typing import Union, Callable
from copy import deepcopy
from ..agents.agent import Agent
from ..problems.problem import Problem
from .algorithm import Algorithm
from ..functions import Variable


class DualDecomposition(Algorithm):
    """Distributed dual decomposition.

    From the perspective of agent :math:`i` the algorithm works as follows.

    Initialization: :math:`\lambda_{ij}^0` for all :math:`j \in \mathcal{N}_i`
    
    For :math:`k=0,1,\\dots`

    * Gather :math:`\lambda_{ji}^k` from neighbors :math:`j \in \mathcal{N}_i`

    * Compute :math:`x_i^k` as an optimal solution of

    .. math::
        \min_{x_i \in X_i} \: f_i(x_i) + x_i^\\top \sum_{j \in \mathcal{N}_i} (\lambda_{ij}^k - \lambda_{ji}^k)
    
    * Gather :math:`x_j^k` from neighbors :math:`j \in \mathcal{N}_i`
    
    * Update for all :math:`j \in \mathcal{N}_i`
    
    .. math::
        \lambda_{ij}^{k+1} = \lambda_{ij}^{k} + \\alpha^k (x_i^k - x_j^k)

    where :math:`x_i\\in\\mathbb{R}^{n}`, :math:`\lambda_{ij}\\in\\mathbb{R}^n` for all :math:`j \in \mathcal{N}_i`, :math:`X_i\\subseteq\mathbb{R}^{n}` for all :math:`i` and :math:`\\alpha^k` is a positive stepsize.

    The algorithm has been presented in ????.
    """

    # TODO choose ref
    def __init__(self, agent: Agent, initial_condition: dict, enable_log: bool = False):
        super(DualDecomposition, self).__init__(agent, enable_log)

        if not isinstance(agent.problem, Problem):
            raise TypeError("The agent must be equipped with a Problem")
        
        if sum(1 for i in agent.problem.objective_function.input_shape if i > 1) > 1:
            raise ValueError("Currently only mono-dimensional objective functions are supported")
        
        if not all([isinstance(x, np.ndarray) for _, x in initial_condition.items()]):
            raise TypeError("The initial condition dictionary can only contain numpy vectors")
        
        if sorted(agent.in_neighbors) != sorted([x for x in initial_condition]):
            raise TypeError("The initial condition dictionary must contain exactly one vector per neighbor")

        # shape of local variable
        self.x_shape = agent.problem.objective_function.input_shape

        # initialize dual variables and primal solution
        self.lambd0 = deepcopy(initial_condition)
        self.lambd = deepcopy(initial_condition)
        self.x = None

    def run(self, iterations: int = 1000, stepsize: Union[float, Callable] = 0.1, verbose: bool=False, **kwargs) -> np.ndarray:
        """Run the algorithm for a given number of iterations

        Args:
            iterations: Number of iterations. Defaults to 1000.
            stepsize: If a float is given as input, the stepsize is constant. 
                                                         If a function is given, it must take an iteration k as input and output the corresponding stepsize.. Defaults to 0.1.
            verbose: If True print some information during the evolution of the algorithm. Defaults to False.
            
        Raises:
            TypeError: The number of iterations must be an int
            TypeError: The stepsize must be a float or a callable

        Returns:
            return a tuple (x, lambda) with the sequence of primal solutions and dual variables if enable_log=True.
        """
        if not isinstance(iterations, int):
            raise TypeError("The number of iterations must be an int")
        if not (isinstance(stepsize, float) or callable(stepsize)):
            raise TypeError("The stepsize must be a float or a function")

        if self.enable_log:
            # initialize sequence of x
            x_dims = [iterations]
            for dim in self.x_shape:
                x_dims.append(dim)
            self.x_sequence = np.zeros(x_dims)

            # initialize sequence of lambda
            self.lambda_sequence = {}
            for j in self.agent.in_neighbors:
                self.lambda_sequence[j] = np.zeros(x_dims)

        for k in range(iterations):
            # store current lambda
            if self.enable_log:
                for j in self.agent.in_neighbors:
                    self.lambda_sequence[j][k] = self.lambd[j]

            # perform an iteration
            if not isinstance(stepsize, float):
                step = stepsize(k)
            else:
                step = stepsize
            self.iterate_run(stepsize=step, **kwargs)

            # store primal solution
            if self.enable_log:
                self.x_sequence[k] = self.x
            
            if verbose:
                if self.agent.id == 0:
                    print('Iteration {}'.format(k), end="\r")

        if self.enable_log:
            return (self.x_sequence, self.lambda_sequence)
    
    def _update_local_solution(self, x: np.ndarray, x_neigh: dict, stepsize: float, **kwargs):
        """Update the local solution
        
        Args:
            x: current solution
            x_neigh: solutions of neighbors (dictionary)
            stepsize: step-size for update
        """
        for j, x_j in x_neigh.items():
            self.lambd[j] += stepsize * (x - x_j)
        
        self.x = x
    
    def iterate_run(self, stepsize: float, **kwargs):
        """Run a single iterate of the algorithm
        """
        # TODO extend to non mono-dimensional variables

        # exchange dual variables with neighbors
        lambda_neigh = self.agent.neighbors_exchange(self.lambd, dict_neigh=True)
        deltalambda = np.zeros(self.x_shape)

        for j in self.agent.in_neighbors:
            deltalambda += self.lambd[j] - lambda_neigh[j]
        
        # build local problem
        x = Variable(self.x_shape[0])
        obj_function = self.agent.problem.objective_function + deltalambda @ x
        pb = Problem(obj_function, self.agent.problem.constraints)

        # solve problem and save data
        x = pb.solve()

        # exchange primal variables with neighbors
        x_neigh = self.agent.neighbors_exchange(x)

        self._update_local_solution(x, x_neigh, stepsize, **kwargs)
    
    def get_result(self):
        """Return the current value primal solution and dual variable
    
        Returns:
            tuple (primal, dual): value of primal solution (np.ndarray), dual variables (dictionary of np.ndarray)
        """
        return (self.x, self.lambd)


from mpi4py import MPI
from threading import Event
from typing import List, Dict, Any
import dill


class Communicator():
    """Communicator abstract class

    """

    def __init__(self):
        pass

    def neighbors_send(self, obj: Any, neighbors: List[int]):
        pass

    def neighbors_receive(self, neighbors: List[int], stop_event: Event) -> Dict[int, Any]:
        pass

    def neighbors_receive_asynchronous(self, neighbors: List[int]) -> Dict[int, Any]:
        pass

    def neighbors_exchange(self, send_obj: Any, in_neighbors: List[int], out_neighbors: List[int],
        dict_neigh: bool, stop_event: Event) -> Dict[int, Any]:
        pass


class MPICommunicator(Communicator):
    """Communicator class that performs communications through MPI. Requires mpi4py.

    Attributes:
        comm: communication world
        size (int): size of the network. 
        rank (int): rank of the processor

    """

    def __init__(self):
        self.comm = MPI.COMM_WORLD
        self.size = MPI.COMM_WORLD.Get_size()
        self.rank = MPI.COMM_WORLD.Get_rank()
        self.requests = []

    def neighbors_send(self, obj: Any, neighbors: List[int]):
        """Send data to neighbors.

        Args:
            obj: object to send
            neighbors: list of out-neighbors
        """
        obj_send = dill.dumps(obj)
        for neighbor in neighbors:
            req = self.comm.Isend(obj_send, dest=neighbor, tag=neighbor)
            self.requests.append(req) # TODO in case this function is called directly, this list may explode

    def neighbors_receive(self, neighbors: List[int], stop_event: Event = None) -> Dict[int, Any]:
        """Receive data from neighbors (waits until data are received from all neighbors).

        Args:
            neighbors: list of in-neighbors
            stop_event: an Event object that is monitored during the execution.
                If the event is set, the function returns immediately.
                Defaults to None (does not wait upon any event)

        Returns:
            data received by in-neighbors
        """
        received_data = {}

        while(len(received_data) < len(neighbors)):

            if stop_event is not None and stop_event.is_set():
                break

            # cycle over remaining neighbors
            for node in [k for k in neighbors if k not in received_data]:
                state = MPI.Status()
                okay = self.comm.Iprobe(source=node, tag=MPI.ANY_TAG, status=state)

                if(okay):
                    data = bytearray(state.Get_count())
                    self.comm.Recv(data, source=node, tag=state.Get_tag())
                    data = dill.loads(data)
                    received_data[node] = data

        return received_data

    def neighbors_receive_asynchronous(self, neighbors: List[int]) -> Dict[int, Any]:
        """Receive data (if any) from neighbors.

        Args:
            neighbors: list of in-neighbors

        Returns:
            data received by in-neighbors (if any)
        """
        received_data = {}
        state = MPI.Status()
        while self.comm.Iprobe(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=state):
            node = state.Get_source()
            data = self.comm.recv(source=node)
            received_data[node] = data
            state = MPI.Status()

        return received_data

    def neighbors_exchange(self, send_obj: Any, in_neighbors: List[int], out_neighbors: List[int],
            dict_neigh: bool=False, stop_event: Event = None) -> Dict[int, Any]:
        """Exchange information (synchronously) with neighbors.

        Args:
            send_obj: object to send
            in_neighbors: list of in-neighbors
            out_neighbors: list of out-neighbors
            dict_neigh: True if send_obj contains a dictionary with different objects for each neighbor.
                Defaults to False
            stop_event: an Event object that is monitored during the execution.
                If the event is set, the function returns immediately.
                Defaults to None (does not wait upon any event)

        Returns:
            data received by in-neighbors
        """

        self.requests = []
        if dict_neigh == False:
            self.neighbors_send(send_obj, out_neighbors)
        else:
            for j in out_neighbors:
                self.neighbors_send(send_obj[j], [j])
        data = self.neighbors_receive(in_neighbors, stop_event)
        MPI.Request.Waitall(self.requests)
        # self.comm.Barrier()
        return data



import numpy as np
from typing import Union, Callable
from threading import Event
from ..agents.agent import Agent
from ..problems.problem import Problem
from ..problems.constraint_coupled_problem import ConstraintCoupledProblem
from .algorithm import Algorithm
from ..functions import Variable
from ..functions import ExtendedFunction
from ..constraints import ExtendedConstraint


class PrimalDecomposition(Algorithm):
    """Distributed primal decomposition.

    From the perspective of agent :math:`i` the algorithm works as follows.

    Initialization: :math:`y_i^0` such that :math:`\sum_{i=1}^N y_i^0 = 0`
    
    For :math:`k=0,1,\\dots`

    * Compute :math:`((x_i^k, \\rho_i^k), \mu_i^k)` as a primal-dual optimal solution of

    .. math::
        :nowrap:

        \\begin{split}
        \min_{x_i, \\rho_i} \hspace{1.1cm} & \: f_i(x_i) + M \\rho_i \\\\
        \mathrm{subj. to} \: \mu_i: \: & \: g_i(x_i) \le y_i^k + \\rho_i \\boldsymbol{1} \\\\
        & \: x_i \in X_i, \\rho_i \ge 0
        \\end{split}
    
    * Gather :math:`\mu_j^k` from :math:`j \in \mathcal{N}_i` and update
    
    .. math::
        y_i^{k+1} = y_i^{k} + \\alpha^k \sum_{j \in \mathcal{N}_i} (\mu_i^k - \mu_j^k)

    where :math:`x_i\\in\\mathbb{R}^{n_i}`, :math:`\mu_i,y_i\\in\\mathbb{R}^S`, :math:`X_i\\subseteq\mathbb{R}^{n_i}` for all :math:`i` and :math:`\\alpha^k` is a positive stepsize.

    The algorithm has been presented in ????.
    """

    # TODO choose ref
    def __init__(self, agent: Agent, initial_condition: np.ndarray, enable_log: bool = False):
        super(PrimalDecomposition, self).__init__(agent, enable_log)

        if not isinstance(agent.problem, ConstraintCoupledProblem):
            raise TypeError("The agent must be equipped with a ConstraintCoupledProblem")
        
        if sum(1 for i in agent.problem.objective_function.input_shape if i > 1) > 1:
            raise ValueError("Currently only mono-dimensional objective functions are supported")

        if sum(1 for i in initial_condition.shape if i > 1) > 1:
            raise ValueError("Currently only mono-dimensional outputs for coupling functions are supported")

        # shape of local variable and coupling constraints
        self.x_shape = agent.problem.objective_function.input_shape
        self.S = initial_condition.size # TODO extend to non mono-dimensional case

        # initialize allocation, primal solution and cost
        self.y0 = np.copy(initial_condition)
        self.y = np.copy(initial_condition)
        self.y_avg = np.copy(initial_condition)
        self.x = None
        self.J = None

        # extended versions of objective function, coupling function and constraints (+ 1 variable for rho)
        self.objective_function = ExtendedFunction(agent.problem.objective_function)
        self.coupling_function = ExtendedFunction(agent.problem.coupling_function)
        self.local_constraints = ExtendedConstraint(agent.problem.constraints)

    def run(self, iterations: int = 1000, stepsize: Union[float, Callable] = 0.1, M: float = 1000.0,
        verbose: bool=False, callback_iter: Callable=None, compute_runavg: bool=False, runavg_start_iter: int=0,
        event: Event=None, **kwargs) -> np.ndarray:
        """Run the algorithm for a given number of iterations

        Args:
            iterations: Number of iterations. Defaults to 1000.
            stepsize: If a float is given as input, the stepsize is constant. 
                                                         If a function is given, it must take an iteration k as input and output the corresponding stepsize.. Defaults to 0.1.
            M: Value of the parameter :math:`M`. Defaults to 1000.
            verbose: If True print some information during the evolution of the algorithm. Defaults to False.
            callback_iter: callback function to be called at the end of each iteration. Must take an iteration k as input. Defaults to None.
            compute_runavg: whether or not to compute also running average of allocation. Defaults to False.
            runavg_start_iter: specifies when to start computing running average (applies only if compute_runavg = True). Defaults to 0.

        Raises:
            TypeError: The number of iterations must be an int
            TypeError: The stepsize must be a float or a callable
            TypeError: The parameter M must be a float

        Returns:
            return a tuple (x, y, J) with the sequence of primal solutionsm allocation estimates and cost if enable_log=True. If compute_runavg=True, then return (x, y, y_avg, J)
        """
        if not isinstance(iterations, int):
            raise TypeError("The number of iterations must be an int")
        if not (isinstance(stepsize, float) or callable(stepsize)):
            raise TypeError("The stepsize must be a float or a function")
        if not isinstance(M, float):
            raise TypeError("The parameter M must be a float")
        if callback_iter is not None and not callable(callback_iter):
            raise TypeError("The callback function must be a Callable")
        if runavg_start_iter < 0:
            raise ValueError("The parameter runavg_start_iter must not be negative")

        if self.enable_log:
            # initialize sequence of costs
            x_dims = [iterations]
            self.J_sequence = np.zeros(x_dims)

            # initialize sequence of x
            for dim in self.x_shape:
                x_dims.append(dim)
            self.x_sequence = np.zeros(x_dims)

            # initialize sequence of y
            y_dims = [iterations]
            for dim in self.y.shape:
                y_dims.append(dim)
            self.y_sequence = np.zeros(y_dims)

            # initialize sequence of averaged y
            if compute_runavg:
                self.y_avg_sequence = np.zeros(y_dims)

        # initialize cumulative sum of stepsize if needed
        if compute_runavg:
            self.stepsize_sum = 0
        
        last_iter = np.copy(iterations)

        for k in range(iterations):
            # store current allocation
            if self.enable_log:
                self.y_sequence[k] = self.y

                if compute_runavg:
                    self.y_avg_sequence[k] = self.y_avg

            # compute stepsize
            if not isinstance(stepsize, float):
                step = stepsize(k)
            else:
                step = stepsize

            # determine whether or not running average must be updated
            update_runavg = compute_runavg and k >= runavg_start_iter

            # perform iteration
            self.iterate_run(stepsize=step, M=M, update_runavg=update_runavg, event=event, **kwargs)

            # store primal solution and cost
            if self.enable_log:
                self.x_sequence[k] = self.x
                self.J_sequence[k] = self.J
            
            # check if we must stop for external event
            if event is not None and event.is_set():
                last_iter = k
                break
            
            # perform external callback (if any)
            if callback_iter is not None:
                callback_iter(k)
            
            if verbose:
                if self.agent.id == 0:
                    print('Iteration {}'.format(k), end="\r")

        if self.enable_log and compute_runavg:
            return (self.x_sequence.take(np.arange(0,last_iter), axis=0), self.y_sequence.take(np.arange(0,last_iter), axis=0),
                self.y_avg_sequence.take(np.arange(0,last_iter), axis=0), self.J_sequence[0:last_iter])
        elif self.enable_log:
            return (self.x_sequence.take(np.arange(0,last_iter), axis=0), self.y_sequence.take(np.arange(0,last_iter), axis=0),
                self.J_sequence[0:last_iter])
    
    def _update_local_solution(self, x: np.ndarray, mu: np.ndarray, mu_neigh: list, stepsize: float, update_runavg, **kwargs):
        """Update the local solution
        
        Args:
            x: current primal solution
            mu: current dual solution
            mu_neigh: dual solutions of neighbors
            stepsize: step-size for update
        """

        y_new = self.y

        for mu_j in mu_neigh:
            y_new += stepsize * (mu - mu_j)
        
        self.x = x
        self.J = np.asscalar(self.agent.problem.objective_function.eval(x))
        self.y = y_new

        if update_runavg:
            self.stepsize_sum += stepsize
            self.y_avg += stepsize * (self.y - self.y_avg) / self.stepsize_sum
    
    def iterate_run(self, stepsize: float, M: float, update_runavg, event: Event, **kwargs):
        """Run a single iterate of the algorithm
        """
        # TODO extend this function to non mono-dimensional case

        # build local problem
        z = Variable(self.x_shape[0] + 1)
        A = np.hstack((np.zeros((self.S, self.x_shape[0])), np.ones((self.S,1)))).transpose()
        A_rho = np.hstack((np.zeros((1, self.x_shape[0])), [[1]])).transpose()
        rho = A_rho @ z
        alloc_constr = self.coupling_function <= self.y + A @ z
        rho_constr = rho >= 0
        constraints = [alloc_constr, rho_constr]
        constraints.extend(self.local_constraints)
        obj_function = self.objective_function + M * rho
        pb = Problem(obj_function, constraints)

        # solve problem and save data
        out = pb.solve(return_only_solution=False)
        if out['status'] != 'solved':
            raise ValueError("The local problem could not be solved")

        x = out['solution'][:-1]
        # rho = out['solution'][-1]
        mu = out['dual_variables'][0]

        # exchange dual variables with neighbors
        data = self.agent.neighbors_exchange(mu, event=event)
        mu_neigh = [data[idx] for idx in data]

        if event is None or not event.is_set():
            self._update_local_solution(x, mu, mu_neigh, stepsize, update_runavg, **kwargs)
    
    def get_result(self, return_runavg:bool = False):
        """Return the current value of primal solution, allocation and cost

        Args:
            return_runavg: whether or not to return also running average of allocation. Defaults to False.
    
        Returns:
            tuple (primal, allocation, cost) of numpy.ndarray: value of primal solution, allocation, cost (if return_runavg = False)
            tuple (primal, allocation, allocation_avg cost) if return_runavg = True
        """
        if return_runavg:
            return (self.x, self.y, self.y_avg, self.J)
        else:
            return (self.x, self.y, self.J)


import numpy as np
import time
from typing import Any 
from ..agents import Agent
from .algorithm import Algorithm


class LogicAnd(Algorithm):
    """Logic-And algorithm. It can be used for checking in a distributed way if a certain condition (corresponding to flag=True in the algorithm) is satisfied by all the agents in the network. Details can be found in [FaGa19a]_
        
    Args:
        agent (Agent): Agent
        graph_diameter (int): diameter of the graph representing the network 
        flag (bool, optional): local flag value. Defaults to False.
        enable_log (bool, optional): True for enabling log. Defaults to False.
    """

    def __init__(self, agent: Agent, graph_diameter: int, flag: bool = False, enable_log: bool = False, **kwargs):
        super(LogicAnd, self).__init__(agent, enable_log, **kwargs)

        self.flag = flag
        self.graph_diameter = graph_diameter
        self.S = np.zeros([graph_diameter, len(self.agent.in_neighbors) + 1])
        self.S_ind = {}
        for j in range(len(self.agent.in_neighbors)):
            self.S_ind[self.agent.in_neighbors[j]] = j

    def change_flag(self, new_flag: bool):
        """Change the local flag

        Args:
            new_flag: new flag
        """
        if not isinstance(new_flag, bool):
            raise TypeError("new flag must be a bool")
        self.flag = new_flag

    def matrix_update(self):
        """Update the matrix S
        """
        self.S[0, -1] = int(self.flag)
        for l in range(self.graph_diameter-1):
            self.S[l+1, -1] = np.prod(self.S[l, :])

    def force_matrix_update(self):
        """
        Force the matrix S to have all ones in the last row
        """
        self.S[-1] = 1

    def update_column(self, neighbor: Any, column: np.ndarray):
        """Update a column of the matrix corresponding to a neighbor
        
        Args:
            neighbor: neighbor
            column: column value
        
        Raises:
            TypeError: second argument must be a numpy.ndarray with shape (graph_diameter, )
            ValueError: second argument must be a numpy.ndarray with shape (graph_diameter, )
        """
        if not isinstance(column, np.ndarray):
            raise TypeError("second argument must be a numpy.ndarray with shape (graph_diameter, )")
        if column.shape != (self.graph_diameter, ):
            raise ValueError("second argument must be a numpy.ndarray with shape (graph_diameter, )")

        index = self.S_ind[neighbor]
        self.S[:, index] = column

    def check_stop(self):
        """Check the last row of S
        
        Returns:
            bool: True if last row contains only ones. Meaning that all have the flag True
        """
        return bool(np.prod(self.S[-1, :]))

    def iterate_run(self):
        """Run an iterate
        """
        data = self.agent.neighbors_exchange(self.S[:, -1])
        for neigh in data:
            self.update_column(neigh, data[neigh])
        self.matrix_update()

    def run(self, maximum_iterations: int = 100, verbose: bool=False):
        """Run the algorithm
        
        Args:
            maximum_iterations: Maximum number of iterations. Defaults to 100.
            verbose: If True print some information during the evolution of the algorithm. Defaults to False.
        
        Raises:
            TypeError: maximum iterations must be an int
        """
        if not isinstance(maximum_iterations, int):
            raise TypeError("maximum iterations must be an int")
        k = 0
        while k < maximum_iterations:
            self.iterate_run()
            if self.check_stop():
                print("logic-and completed in {} iterations".format(k))
                break
            if verbose:
                if self.agent.id == 0:
                    print('Iteration {}'.format(k), end="\r")
            k += 1
    
    def matrix_reset(self):
        self.flag = False
        self.S = np.zeros(self.S.shape)


class AsynchronousLogicAnd(LogicAnd):
    """Asyncrhonous Logic-And algorithm. It can be used for checking in a distributed way if a certain condition (corresponding to flag=True in the algorithm) is satisfied by all the agents in the network. Details can be found in [FaGa19a]_
        
    Args:
        agent (Agent): Agent
        graph_diameter (int): diameter of the graph representing the network 
        flag (bool, optional): local flag value. Defaults to False.
        enable_log (bool, optional): True for enabling log. Defaults to False.
    """
    def iterate_run(self):
        """Run an iterate
        """
        data = self.agent.neighbors_receive_asynchronous()

        for neigh in data:
            self.update_column(neigh, data[neigh])
        self.matrix_update()

        self.agent.neighbors_send(self.S[:, -1])

    def run(self, maximum_running_time: float = 1):
        """Run the algorithm
        
        Args:
            maximum_running_time: Maximum running time. Defaults to 1.
        
        Raises:
            TypeError: maximum running time must be a float
        """
        if not isinstance(maximum_running_time, float):
            raise TypeError("maximum running time must be a float")
        start_time = time.time()
        end_time = start_time + maximum_running_time
        while time.time() <= end_time:
            self.iterate_run()
            if self.check_stop():
                print("logic-and completed in {} s".format(time.time()-start_time))
                break

class MaxConsensus(Algorithm):
    """Max-Consensus algorithm. It computes the entry-wise maximum of a numpy array by using only neighboring communication.
        
    Args:
        agent (Agent): Agent
        x0 (np.ndarray): local initial condition
        graph_diameter (int, optional): diameter of the graph representing the network
        enable_log (bool, optional): True to enable log. Defaults to False.
    """

    def __init__(self, agent: Agent, x0: np.ndarray, graph_diameter: int = None, enable_log: bool = False, **kwargs):
        super(MaxConsensus, self).__init__(agent, enable_log, **kwargs)

        self.x0 = x0
        self.x  = x0
        self.graph_diameter = graph_diameter
        self.stop_iterations = None

        # if the graph diameter is provided, set stopping criterion
        if graph_diameter is not None:
            self.stop_iterations = 2 * graph_diameter + 1
    
    def iterate_run(self):
        """Run an iterate
        """
        data = self.agent.neighbors_exchange(self.x)

        for neigh in data:
            self.x = np.maximum(self.x, data[neigh])

    def run(self, iterations: int = 100, verbose: bool=False):
        """Run the algorithm
        
        Args:
            iterations: Maximum number of iterations. Defaults to 100.
            verbose: If True print some information during the evolution of the algorithm. Defaults to False.
        
        Raises:
            TypeError: maximum iterations must be an int
        """
        if not isinstance(iterations, int) or iterations <= 0:
            raise TypeError("iterations must be a positive integer")

        if self.enable_log:
            # initialize sequence
            dims = [iterations]
            dims.extend(self.x.shape)
            self.sequence_x = np.zeros(dims)
        
        # initialize counter for stopping criterion
        counter = 0
        last_iter = np.copy(iterations)

        for k in range(iterations):
            # store previous value
            prev_x = self.x

            # perform an iteration
            self.iterate_run()

            # store solution sequence
            if self.enable_log:
                self.sequence_x[k] = self.x

            # print information
            if verbose and self.agent.id == 0:
                print('Iteration {}'.format(k), end="\r")

            # increase counter if solution has not changed, otherwise reset
            if np.linalg.norm(prev_x - self.x) < 1e-6:
                counter += 1
            else:
                counter = 0
            
            # check termination condition
            if self.stop_iterations is not None and counter > self.stop_iterations:
                # convergence detected
                self.agent.neighbors_send(self.x) # broadcast local basis a last time
                last_iter = k+1
                break

        # return sequences
        if self.enable_log:
            return self.sequence_x.take(np.arange(0,last_iter), axis=0)
        
    def get_result(self):
        return self.x


import unittest
import numpy as np
import warnings

from ..functions import *

class TestFunction(unittest.TestCase):
    """unit test for functions module"""

    def setUp(self):
        self.n = 3
        self.m = 2

        self.x = Variable(self.n)

        P = np.random.randn(self.n, self.n)
        self.P = P.transpose() @ P
        self.q = np.random.rand(self.n, 1)

        self.A = np.random.randn(self.n, self.m)
        self.b = np.random.rand(self.m, 1)
    
    def test_affine(self):
        warnings.simplefilter('ignore', category=UserWarning)
        affine_form = self.A @ self.x + self.b
        affine_form_implicit = AffineForm(self.x, self.A, self.b)

        self.assertEqual(affine_form, affine_form_implicit)
        pt = np.random.randn(*self.x.input_shape)
        np.testing.assert_almost_equal(affine_form.jacobian(pt), affine_form._alternative_jacobian(pt))
        self.assertIsInstance(affine_form @ affine_form, QuadraticForm)

if __name__ == '__main__':
    unittest.main()


import numpy as np
from .problem import Problem
from ..functions import AffineForm
from ..functions.affine_form import aggregate_affine_form
from ..constraints import Constraint, AbstractSet
from .utilities import check_affine_constraints


class LinearProblem(Problem):
    """Solve a Linear programming problem defined as:

    .. math::

        \\text{minimize } & c^\\top x

        \\text{subject to } & G x \\leq h

                            & A x = b
    """

    def __new__(cls, *args, **kwargs):
        instance = object.__new__(cls)
        return instance

    def __init__(self, objective_function: AffineForm, constraints: list = None, **kwargs):
        """[summary]

        Args:
            objective_function (AffineForm): [description]
            constraints (list, optional): [description]. Defaults to None.
        """
        self.objective_function = None
        self.aggregated_constraints = []
        self.constraints = []
        self.input_shape = None
        self.output_shape = None
        self.set_objective_function(objective_function)
        if constraints is not None:
            if not check_affine_constraints(constraints):
                raise ValueError("Constraints must be affine")
            self.__set_constraints(constraints)

    def set_objective_function(self, objective_function: AffineForm):
        """set the objective function

        Args:
            objective_function (AffineForm): objective function

        Raises:
            TypeError: Objective function must be a AffineForm with output_shape=(1,1)
        """
        if not isinstance(objective_function, AffineForm):
            raise TypeError("Objective function must be a AffineForm")
        if not objective_function.output_shape == (1,1):
            raise TypeError("Objective function must be a AffineForm with output_shape=(1,1)")

        self.objective_function = objective_function
        self.input_shape = objective_function.input_shape
        self.output_shape = objective_function.output_shape

    def __set_constraints(self, constraints):
        """Set the constraints

        Args:
            constraints (list): list of constraints

        Raises:
            TypeError: a list of affine Constraints must be provided
        """
        if not isinstance(constraints, list):
            constraints = [constraints]
        for constraint in constraints:
            super().add_constraint(constraint)

        self.constraints_map = {}
        equality_idx_counter = 0
        inequality_idx_counter = 0
        equalities = []
        inequalities = []
        for idx, constraint in enumerate(self.constraints):
            if constraint.is_affine:
                if constraint.is_equality:
                    equalities.append(constraint.function)
                    self.constraints_map[idx] = {'type': 'affine_equality', 'indices': list(
                        range(equality_idx_counter, equality_idx_counter+constraint.output_shape[0]))}
                    equality_idx_counter += constraint.output_shape[0]
                elif constraint.is_inequality:
                    inequalities.append(constraint.function)
                    self.constraints_map[idx] = {'type': 'affine_inequality', 'indices': list(
                        range(inequality_idx_counter, inequality_idx_counter+constraint.output_shape[0]))}
                    inequality_idx_counter += constraint.output_shape[0]
            else:
                raise TypeError("a list of affine Constraints must be provided")

        if len(equalities) != 0:
            aggregated_equality = aggregate_affine_form(equalities)
            self.aggregated_constraints.append(aggregated_equality == 0)

        if len(inequalities) != 0:
            aggregated_inequality = aggregate_affine_form(inequalities)
            self.aggregated_constraints.append(aggregated_inequality <= 0)

    # def add_constraint(self, constraint):
    #     """A a new (affine) constraint

    #     Args:
    #         constraint (Constraint): affine constraint

    #     Raises:
    #         TypeError: an affine Constraints must be provided
    #     """
    #     if not isinstance(constraint, (Constraint, AbstractSet)):
    #         raise TypeError("an affine Constraints must be provided")
    #     if constraint.is_affine:
    #         if constraint.is_equality:
    #             if len(self.constraints) == 0:
    #                 self.constraints.append(constraint)
    #             else:
    #                 equality_found = False
    #                 for const in self.constraints:
    #                     if const.is_equality:
    #                         equality_found = True
    #                         aggregated_equality = aggregate_affine_form([const.function, constraint.function])
    #                         self.constraints.remove(const)
    #                         self.constraints.append(aggregated_equality == 0)

    #                 if not equality_found:
    #                     self.constraints.append(constraint)
    #         elif constraint.is_inequality:
    #             if len(self.constraints) == 0:
    #                 self.constraints.append(constraint)
    #             else:
    #                 inequality_found = False
    #                 for const in self.constraints:
    #                     if const.is_inequality:
    #                         inequality_found = True
    #                         aggregated_inequality = aggregate_affine_form([const.function, constraint.function])
    #                         self.constraints.remove(const)
    #                         self.constraints.append(aggregated_inequality <= 0)

    #                 if not inequality_found:
    #                     self.constraints.append(constraint)
    #     else:
    #         raise TypeError("an affine Constraints must be provided")

    def __get_cvxopt_objective_parameters(self):
        """Return the objective parameter required by cvxopt

        Returns:
            numpy.ndarray
        """
        c,  _ = self.objective_function.get_parameters()
        return c

    def __get_cvxopt_constraints_parameters(self):
        """Return the constraints parameters required by cvxopt

        Returns:
            tuple: (A, b, G, h)
        """
        A = None
        b = None
        G = None
        h = None
        if len(self.aggregated_constraints) != 0:
            for const in self.aggregated_constraints:
                if const.is_equality:
                    A, b = const.get_parameters()
                    A = A.transpose()
                    b = -b
                if const.is_inequality:
                    G, h = const.get_parameters()
                    G = G.transpose()
                    h = -h

        return G, h, A, b

    def __solve_cvxopt(self, initial_value: np.ndarray = None, solver='glpk', return_only_solution: bool = True):
        """Solve the problem through cvxopt interface

        Args:
            initial_value (np.ndarray, optional): starting point for warm start (if possible). Defaults to None.
            solver (str, optional): ('glpk' or 'cvxopt'). Defaults to 'glpk'.

        Raises:
            ValueError: solver must be 'glpk' or 'cvxopt'
            ValueError: optimal solution not found

        Returns:
            numpy.ndarray: solution
        """
        try:
            import cvxopt
        except ImportError:
            raise ImportError("CVXOPT is required")
        cvxopt.solvers.options['show_progress'] = False
        cvxopt.solvers.options['maxiters'] = 10000
        if solver == 'glpk':
            try:
                import cvxopt.glpk
                solver = 'glpk'
                cvxopt.solvers.options['glpk'] = {'msg_lev': 'GLP_MSG_OFF'}
            except ImportError:
                print("GLPK solver not found. Trying with cvxopt...")
                solver = None
        elif solver == 'cvxopt':
            solver = None
        else:
            raise ValueError("Unsupported solver")

        c = self.__get_cvxopt_objective_parameters()
        G, h, A, b = self.__get_cvxopt_constraints_parameters()
        args = [cvxopt.matrix(c)]
        if G is not None:
            G = cvxopt.matrix(G)
            h = cvxopt.matrix(h)
        else:
            G = cvxopt.matrix(np.ones(c.shape).transpose())
            h = cvxopt.matrix(np.inf)
        args.extend([G, h])
        if A is not None:
            A = cvxopt.matrix(A)
            b = cvxopt.matrix(b)
            args.extend([A, b])
        if initial_value is not None:
            sol = cvxopt.solvers.lp(*args, solver=solver, primal_start=initial_value)
        else:
            sol = cvxopt.solvers.lp(*args, solver=solver)

        if 'optimal' not in sol['status']:
            raise ValueError("LP optimum not found: %s" % sol['status'])

        if not return_only_solution:
            dual_variables = {}
            for idx, constr in enumerate(self.constraints):
                if self.constraints_map[idx]['type'] == 'affine_equality':
                    dual_variables[idx] = np.array(sol['y'])[self.constraints_map[idx]['indices']]
                elif self.constraints_map[idx]['type'] == 'affine_inequality':
                    dual_variables[idx] = np.array(sol['z'])[self.constraints_map[idx]['indices']]

            if sol['status'] == 'optimal':
                status = 'solved'
            else:
                status = 'unknown'

            output_dict = {
                'solution': np.array(sol['x']).reshape(self.input_shape),
                'status': status,
                'dual_variables': dual_variables
            }
            return output_dict
        else:
            return np.array(sol['x']).reshape(self.input_shape)

    def solve(self, initial_value: np.ndarray = None, solver='glpk', return_only_solution: bool = True):
        """Solve the problem

        Args:
            initial_value (numpy.ndarray), optional): Initial value for warm start. Defaults to None.
            solver (str, optional): Solver to use ['glpk', 'cvxopt']. Defaults to 'glpk'.

        Raises:
            ValueError: Unsupported solver

        Returns:
            numpy.ndarray: solution
        """
        if solver in ['glpk', 'cvxopt']:
            return self.__solve_cvxopt(initial_value=initial_value, solver=solver, return_only_solution=return_only_solution)
        else:
            raise ValueError("Unsupported solver")



import numpy as np
import scipy.sparse as sp
import warnings
from .problem import Problem
from ..functions import AffineForm, QuadraticForm
from ..functions.affine_form import aggregate_affine_form
from ..constraints import Constraint
from ..utils.utilities import check_symmetric, is_semi_pos_def
from .utilities import check_affine_constraints


class QuadraticProblem(Problem):
    """Solve a Quadratic programming problem defined as:

    .. math::

        \\text{minimize } & x^\\top P x + q^\\top x + r 

        \\text{subject to } & G x \\leq h

                            & A x = b

    Quadratic problems are currently solved by using CVXOPT or OSQP https://osqp.org. 

    Args:
        objective_function (QuadraticForm): objective function
        constraints (list, optional): list of constraints. Defaults to None.
        is_pos_def (bool): True if P is (semi)positive definite. Defaults to True.
    """

    def __new__(cls, *args, **kwargs):
        instance = object.__new__(cls)
        return instance

    def __init__(self, objective_function: QuadraticForm, constraints: list = None, is_pos_def: bool = True, **kwargs):
        self.objective_function = None
        self.constraints = []
        self.input_shape = None
        self.output_shape = None
        self.is_pos_def = is_pos_def
        self.set_objective_function(objective_function)
        if not check_affine_constraints(constraints):
            raise ValueError("Constraints must be affine")
        self.aggregated_constraints = []
        if constraints is not None:
            self.set_constraints(constraints)
        # TODO: check init is done twice if instantiated via Problem

    def set_objective_function(self, objective_function: QuadraticForm):
        """set the objective function

        Args:
            objective_function (QuadraticForm): objective function

        Raises:
            TypeError: Objective function must be a QuadraticForm
        """
        if not isinstance(objective_function, QuadraticForm):
            raise TypeError("Objective function must be a QuadraticForm")
        P, _, _ = objective_function.get_parameters()
        if not self.is_pos_def:
            if not is_semi_pos_def(P):
                warnings.warn("Objective function is not convex.")
        self.objective_function = objective_function
        self.input_shape = objective_function.input_shape
        self.output_shape = objective_function.output_shape

    def set_constraints(self, constraints):
        """Set the constraints

        Args:
            constraints (list): list of constraints

        Raises:
            TypeError: a list of affine Constraints must be provided
        """
        if not isinstance(constraints, list):
            constraints = [constraints]
        for constraint in constraints:
            super().add_constraint(constraint)

        self.constraints_map = {}
        equality_idx_counter = 0
        inequality_idx_counter = 0
        equalities = []
        inequalities = []
        for idx, constraint in enumerate(self.constraints):
            if constraint.is_affine:
                if constraint.is_equality:
                    equalities.append(constraint.function)
                    self.constraints_map[idx] = {'type': 'affine_equality', 'indices': list(
                        range(equality_idx_counter, equality_idx_counter+constraint.output_shape[0]))}
                    equality_idx_counter += constraint.output_shape[0]
                elif constraint.is_inequality:
                    inequalities.append(constraint.function)
                    self.constraints_map[idx] = {'type': 'affine_inequality', 'indices': list(
                        range(inequality_idx_counter, inequality_idx_counter+constraint.output_shape[0]))}
                    inequality_idx_counter += constraint.output_shape[0]
            else:
                raise TypeError("a list of affine Constraints must be provided")

        for key, item in self.constraints_map.items():
            if item['type'] == 'affine_inequality':
                item['indices'] = list(np.array(item['indices']) + equality_idx_counter)

        if len(equalities) != 0:
            aggregated_equality = aggregate_affine_form(equalities)
            self.aggregated_constraints.append(aggregated_equality == 0)

        if len(inequalities) != 0:
            aggregated_inequality = aggregate_affine_form(inequalities)
            self.aggregated_constraints.append(aggregated_inequality <= 0)

    # def add_constraint(self, constraint):
    #     """A a new (affine) constraint

    #     Args:
    #         constraint (Constraint): affine constraint

    #     Raises:
    #         TypeError: an affine Constraints must be provided
    #     """
    #     if not isinstance(constraint, Constraint):
    #         raise TypeError("an affine Constraints must be provided")
    #     if constraint.is_affine:
    #         if constraint.is_equality:
    #             if len(self.constraints) == 0:
    #                 self.constraints.append(constraint)
    #             else:
    #                 equality_found = False
    #                 for const in self.constraints:
    #                     if const.is_equality:
    #                         equality_found = True
    #                         aggregated_equality = aggregate_affine_form([const.function, constraint.function])
    #                         self.constraints.remove(const)
    #                         self.constraints.append(aggregated_equality == 0)

    #                 if not equality_found:
    #                     self.constraints.append(constraint)
    #         elif constraint.is_inequality:
    #             if len(self.constraints) == 0:
    #                 self.constraints.append(constraint)
    #             else:
    #                 inequality_found = False
    #                 for const in self.constraints:
    #                     if const.is_inequality:
    #                         inequality_found = True
    #                         aggregated_inequality = aggregate_affine_form([const.function, constraint.function])
    #                         self.constraints.remove(const)
    #                         self.constraints.append(aggregated_inequality <= 0)

    #                 if not inequality_found:
    #                     self.constraints.append(constraint)
    #     else:
    #         raise TypeError("an affine Constraints must be provided")

    def __get_osqp_objective_parameters(self):
        """return the objective parameter fo osqp solver
        """
        P, q, _ = self.objective_function.get_parameters()
        if isinstance(P, np.ndarray):
            if not check_symmetric(P):
                raise ValueError("Matrix P must be symmetric.")
            # if not is_semi_pos_def(P):
            #     raise ValueError("Nonconvex objective function.")
            P = sp.csc_matrix(2*P)
        return P, q.flatten()

    def __get_osqp_constraints_parameters(self):
        """return the constraints parameter fo osqp solver
        """
        A = None
        b = None
        G = None
        h = None
        if len(self.aggregated_constraints) != 0:
            for const in self.aggregated_constraints:
                if const.is_equality:
                    A, b = const.get_parameters()
                if const.is_inequality:
                    G, h = const.get_parameters()

            if A is not None:
                A_qp = A
                l_qp = -b
                u_qp = -b

                if G is not None:
                    A_qp = np.hstack([A, G])
                    l_qp = np.vstack([-b, -np.inf * np.ones((len(h), 1))])
                    u_qp = np.vstack([-b, -h])
            else:
                A_qp = G
                l_qp = -np.inf * np.ones(len(h))
                u_qp = -h

            A_osqp = sp.csc_matrix(A_qp.transpose())
            l_osqp = l_qp.flatten()
            u_osqp = u_qp.flatten()

            return A_osqp, l_osqp, u_osqp
        else:
            return None, None, None

    def __solve_osqp(self, initial_value: np.ndarray = None, return_only_solution: bool = True):
        """return the problem through osqp
        """
        try:
            from osqp import OSQP
        except ImportError:
            raise ImportError("OSQP is not installed")

        P, q = self.__get_osqp_objective_parameters()
        A, l, u = self.__get_osqp_constraints_parameters()
        osqp_pb = OSQP()
        osqp_pb.setup(P=P, q=q, A=A, l=l, u=u, verbose=False)
        osqp_pb.update_settings(eps_abs=1e-7, eps_rel=1e-7, max_iter=10000)

        if initial_value is not None:
            osqp_pb.warm_start(x=initial_value)
        res = osqp_pb.solve()
        if res.info.status != 'solved':
            warnings.warn("OSQP exited with status {}".format(res.info.status))

        solution = res.x.reshape(self.input_shape)
        if not return_only_solution:
            dual_variables = {}
            for idx, _ in enumerate(self.constraints):
                dual_variables[idx] = res.y[self.constraints_map[idx]['indices']].reshape(-1, 1)

            output_dict = {
                'solution': solution,
                'status': res.info.status,
                'dual_variables': dual_variables,
            }
            return output_dict
        else:
            return solution

    def __get_cvxopt_objective_parameters(self):
        """return the objective parameter fo osqp solver
        """
        P, q, _ = self.objective_function.get_parameters()
        if isinstance(P, np.ndarray):
            if not is_semi_pos_def(P):
                raise ValueError("Nonconvex objective function.")
        return 2*P, q

    def __get_cvxopt_constraints_parameters(self):
        """return the constraints parameter fo osqp solver
        """
        A = None
        b = None
        G = None
        h = None
        if len(self.aggregated_constraints) != 0:
            for const in self.aggregated_constraints:
                if const.is_equality:
                    A, b = const.get_parameters()
                    A = A.transpose()
                    b = -b
                if const.is_inequality:
                    G, h = const.get_parameters()
                    G = G.transpose()
                    h = -h
        return A, b, G, h

    def __solve_cvxopt(self, initial_value: np.ndarray = None, return_only_solution: bool = True):
        """return the problem through osqp
        """
        try:
            import cvxopt
        except ImportError:
            raise ImportError("CVXOPT is required")
        cvxopt.solvers.options['show_progress'] = False
        cvxopt.solvers.options['maxiters'] = 10000

        P, q = self.__get_cvxopt_objective_parameters()
        A, b, G, h = self.__get_cvxopt_constraints_parameters()
        if initial_value is not None:
            initvals = {'x': initial_value}
        else:
            initvals = None
        if A is not None:
            A = cvxopt.matrix(A)
            b = cvxopt.matrix(b)
        if G is not None:
            G = cvxopt.matrix(G)
            h = cvxopt.matrix(h)

        sol = cvxopt.solvers.qp(P=cvxopt.matrix(P),
                                q=cvxopt.matrix(q),
                                G=G,
                                h=h,
                                A=A,
                                b=b,
                                initvals=initvals)

        solution = np.array(sol['x']).reshape(self.input_shape)
        if not return_only_solution:
            dual_variables = {}
            for idx, constr in enumerate(self.constraints):
                if constr.is_inequality:
                    dual_variables[idx] = np.array(sol['z'])[self.constraints_map[idx]['indices']]
                elif constr.is_equality:
                    dual_variables[idx] = np.array(sol['y'])[self.constraints_map[idx]['indices']]

            if sol['status'] == 'optimal':
                status = 'solved'
            else:
                status = sol['status']

            output_dict = {
                'solution': solution,
                'status': status,
                'dual_variables': dual_variables,
            }
            return output_dict
        else:
            return solution

    def solve(self, initial_value: np.ndarray = None, solver='osqp', return_only_solution: bool = True):
        """Solve the problem

        Args:
            initial_value (numpy.ndarray), optional): Initial value for warm start. Defaults to None.
            solver (str, optional): Solver to use ('osqp' or 'cvxopt'). Defaults to 'osqp'.

        Raises:
            ValueError: Unsupported solver, only 'osqp' and 'cvxopt' are currently supported

        Returns:
            numpy.ndarray: solution
        """
        P, _, _ = self.objective_function.get_parameters()#self.__get_cvxopt_objective_parameters()
        if solver == 'osqp':
            if not check_symmetric(P):
                solver = 'cvxopt'
                warnings.warn("OSQP solver requires the matrix of the quadratic function to be symmetric. Solving with cvxopt.")
        if solver == 'osqp':
            return self.__solve_osqp(initial_value=initial_value, return_only_solution=return_only_solution)
        elif solver == 'cvxopt':
            try:
                return self.__solve_cvxopt(initial_value=initial_value, return_only_solution=return_only_solution)
            except:
                return super().solve(return_only_solution=return_only_solution)
        else:
            raise ValueError("Unsupported solver")


import numpy as np
import autograd.numpy as anp
import warnings
from .abstract_function import AbstractFunction
from .utilities import check_input


class Power(AbstractFunction):
    """Power function (elementwise)

    .. math::

        f(x)= x^\\alpha

    with :math:`x: \\mathbb{R}^{n}`, :math:`\\alpha: \\mathbb{R}`.

    Args:
        fn (AbstractFunction): input function

    Raises:
        TypeError: input must be a AbstractFunction object
    """

    # TODO Recast as square if exponent is 2
    def __init__(self, fn: AbstractFunction, exponent: float):
        if not isinstance(fn, AbstractFunction):
            raise TypeError("Input must be a AbstractFunction object")

        if not fn.is_differentiable:
            pass
            # warnings.warn(
            #     'Composition with a nondifferentiable function will lead to an\
            #         error when asking for a subgradient')
        else:
            self.differentiable = True

        self.fn = fn
        self.exponent = exponent

        self.input_shape = fn.input_shape
        self.output_shape = fn.output_shape

        self.affine = False
        self.quadratic = False

        super().__init__()
        
    def _expression(self):
        expression = 'Power({}, {})'.format(self.fn._expression(), self.exponent)
        return expression 

    def _to_cvxpy(self):
        import cvxpy as cvx
        return cvx.power(self.fn._to_cvxpy(), p=self.exponent)

    @check_input
    def eval(self, x: np.ndarray) -> np.ndarray:
        return anp.power(self.fn.eval(x), self.exponent).reshape(self.output_shape)
    
    def _extend_variable(self, n_var, axis, pos):
        return Power(self.fn._extend_variable(n_var, axis, pos), self.exponent)


import numpy as np
import warnings
from .abstract_function import AbstractFunction
from .utilities import check_input


class ExtendedFunction(AbstractFunction):
    """Function with extended variable

    .. math::

        f(x, y) = x

    with :math:`x\\in \\mathbb{R}^{n}, y\\in \\mathbb{R}^{m}`

    Args:
        fn (AbstractFunction): input function
        n_var: number of additional variables. Defaults to 1
        axis: axis along which the additional variables are appended. Defaults to -1 (the last valid one)
        pos: position index of the old variable vector. Defaults to 0

    Raises:
        TypeError: fn must be a AbstractFunction
        TypeError: n_var must be a positive int
        TypeError: axis must be int
    """

    # TODO implement __new__ to maintain existing structure (e.g. QuadraticForm, AffineForm)

    def __new__(cls, fn: AbstractFunction, n_var: int=1, axis: int=-1, pos: int=0):
        instance = object.__new__(cls)
        if fn.is_affine:
            # TODO extend to non mono-dimensional case
            from .variable import Variable
            from .affine_form import AffineForm

            A, b = fn.get_parameters()
            A_new = np.zeros((A.shape[0]+n_var, A.shape[1]))
            A_new[pos:pos+A.shape[0], :] = A
            instance = AffineForm(Variable(A.shape[0] + n_var), A_new, b)
        elif fn.is_quadratic:
            # TODO extend to non mono-dimensional case
            from .variable import Variable
            from .quadratic_form import QuadraticForm

            P, q, r = fn.get_parameters()
            P_new = np.pad(P, ((pos, n_var-pos), (pos, n_var-pos)), 'constant') # TODO
            q_new = np.zeros((q.shape[0]+n_var, q.shape[1]))
            q_new[pos:pos+q.shape[0], :] = q
            instance = QuadraticForm(Variable(P.shape[0] + n_var), P_new, q_new, r)
        return instance

    def __init__(self, fn: AbstractFunction, n_var: int=1, axis: int=-1, pos: int=0):
        if not isinstance(fn, AbstractFunction):
            raise TypeError("fn must be a AbstractFunction object")
        
        if not isinstance(n_var, int) or n_var <= 0:
            raise TypeError("n_var must be a positive int")
        
        if not isinstance(axis, int):
            raise TypeError("axis must be int")
        
        # internal variables
        self.n_var = n_var
        self.axis = [i for i in range(len(fn.input_shape)) if fn.input_shape[i] > 1][axis]
        self.pos = pos
        self.fn = fn
        self.input_shape_original = fn.input_shape
        indices = [range(i) for i in fn.input_shape]
        indices[self.axis] = range(pos, pos+fn.input_shape[self.axis])
        self.eval_index = np.ix_(*tuple(indices))

        # AbstractFunction variables
        new_shape = list(fn.input_shape)
        new_shape[self.axis] += n_var
        self.input_shape = tuple(new_shape)
        self.output_shape = fn.output_shape
        self.differentiable = fn.differentiable
        self.affine = False
        self.quadratic = False

        super().__init__()
    
    def _expression(self):
        expression = 'Extended Function({}, {} additional vars along axis {} old position {})'.format(self.fn._expression(), self.n_var, self.axis, self.pos)
        return expression 

    @check_input
    def eval(self, x: np.ndarray) -> np.ndarray:
        return self.fn.eval(x[self.eval_index])
    
    def _to_cvxpy(self):
        import cvxpy as cvx
        return self.fn._extend_variable(self.n_var, self.axis, self.pos)._to_cvxpy()


import numpy as np
from typing import Union, Callable
from ..agents import Agent
from .algorithm import Algorithm
from .consensus import PushSumConsensus


class GradientTracking(Algorithm):
    """Gradient Tracking Algorithm [...]_ 

    From the perspective of agent :math:`i` the algorithm works as follows. For :math:`k=0,1,\\dots`

    .. math::

        x_i^{k+1} & = \\sum_{j=1}^N w_{ij} x_j^k - \\alpha d_i^k \\\\
        d_i^{k+1} & = \\sum_{j=1}^N w_{ij} d_j^k - [ \\nabla f_i (x_i^{k+1}) - \\nabla f_i (x_i^k)]


    where :math:`x_i\\in\\mathbb{R}^n` and :math:`d_i\\in\\mathbb{R}^n`. The weight matrix :math:`W=[w_{ij}]` must be doubly-stochastic. Extensions to other class of weight matrices :math:`W` are not currently supported.

    Args:
        agent (Agent): agent to execute the algorithm
        initial_condition (numpy.ndarray): initial condition for :math:`x_i`
        enable_log (bool): True for enabling log

    Attributes:
        agent (Agent): agent to execute the algorithm
        x0 (numpy.ndarray): initial condition
        x (numpy.ndarray): current value of the local solution
        d (numpy.ndarray): current value of the local tracker
        shape (tuple): shape of the variable
        x_neigh (dict): dictionary containing the local solution of the (in-)neighbors
        d_neigh (dict): dictionary containing the local tracker of the (in-)neighbors
        enable_log (bool): True for enabling log
    """

    # TODO choose ref
    def __init__(self, agent: Agent, initial_condition: np.ndarray, enable_log: bool = False):
        super(GradientTracking, self).__init__(agent, enable_log)

        self.x0 = initial_condition
        self.x = initial_condition
        # Initialize tracker at x0
        self.d = self.agent.problem.objective_function.subgradient(initial_condition)

        self.shape = self.x.shape

        self.x_neigh = {}
        self.d_neigh = {}

    def _update_local_solution(self, x: np.ndarray, **kwargs):
        """update the local solution

        Args:
            x: new value

        Raises:
            TypeError: Input must be a numpy.ndarray
            ValueError: Incompatible shapes
        """
        if not isinstance(x, np.ndarray):
            raise TypeError("Input must be a numpy.ndarray")
        if x.shape != self. x0.shape:
            raise ValueError("Incompatible shapes")
        self.x = x

    def _update_local_tracker(self, d: np.ndarray, **kwargs):
        """update the local tracker

        Args:
            d: new value

        Raises:
            TypeError: Input must be a numpy.ndarray
            ValueError: Incompatible shapes
        """
        if not isinstance(d, np.ndarray):
            raise TypeError("Input must be a numpy.ndarray")
        if d.shape != self. x0.shape:
            raise ValueError("Incompatible shapes")
        self.d = d

    def iterate_run(self, stepsize: float, **kwargs):
        """Run a single iterate of the gradient tracking algorithm
        """

        data_x = self.agent.neighbors_exchange(self.x)
        for neigh in data_x:
            self.x_neigh[neigh] = data_x[neigh]

        data_d = self.agent.neighbors_exchange(self.d)
        for neigh in data_d:
            self.d_neigh[neigh] = data_d[neigh]

        x_kp = self.agent.in_weights[self.agent.id] * self.x
        d_kp = self.agent.in_weights[self.agent.id] * self.d

        for j in self.agent.in_neighbors:
            x_kp += self.agent.in_weights[j] * self.x_neigh[j]
            d_kp += self.agent.in_weights[j] * self.d_neigh[j]

        x_kp += - stepsize * self.d
        d_kp += self.agent.problem.objective_function.subgradient(
            x_kp) - self.agent.problem.objective_function.subgradient(self.x)

        self._update_local_solution(x_kp, **kwargs)
        self._update_local_tracker(d_kp, **kwargs)

    def run(self, iterations: int = 1000, stepsize: Union[float, Callable] = 0.1, verbose: bool=False) -> np.ndarray:
        """Run the gradient tracking algorithm for a given number of iterations

        Args:
            iterations: Number of iterations. Defaults to 1000.
            stepsize: If a float is given as input, the stepsize is constant. 
                                                        Default is 0.01.
            verbose: If True print some information during the evolution of the algorithm. Defaults to False.

        Raises:
            TypeError: The number of iterations must be an int
            TypeError: The stepsize must be a float

        Returns:
            return the sequence of estimates if enable_log=True.
        """
        if not isinstance(iterations, int):
            raise TypeError("The number of iterations must be an int")
        if not (isinstance(stepsize, float) or callable(stepsize)):
            raise TypeError("The stepsize must be a float or a function")

        if self.enable_log:
            dims = [iterations]
            for dim in self.x.shape:
                dims.append(dim)
            self.sequence = np.zeros(dims)

        for k in range(iterations):
            if not isinstance(stepsize, float):
                step = stepsize(k)
            else:
                step = stepsize
            self.iterate_run(stepsize=step)

            if self.enable_log:
                self.sequence[k] = self.x
            
            if verbose:
                if self.agent.id == 0:
                    print('Iteration {}'.format(k), end="\r")

        if self.enable_log:
            return self.sequence

    def get_result(self):
        """Return the actual value of x

        Returns:
            numpy.ndarray: value of x
        """
        return self.x


class DirectedGradientTracking(PushSumConsensus):
    """Gradient Tracking Algorithm [XiKh18]_ 

    From the perspective of agent :math:`i` the algorithm works as follows. For :math:`k=0,1,\\dots`

    .. math::

        x_i^{k+1} & = \\sum_{j=1}^N a_{ij} x_j^k - \\alpha y_i^k \\\\
        y_i^{k+1} & = \\sum_{j=1}^N b_{ij} (y_j^k - [ \\nabla f_j (x_j^{k+1}) - \\nabla f_j (x_j^k)])


    where :math:`x_i\\in\\mathbb{R}^n` and :math:`y_i\\in\\mathbb{R}^n`. The weight matrix :math:`A=[a_{ij}]` must be row-stochastic, while :math:`B=[b_{ij}]` must be column-stochastic. The underlying graph can be directed (and unbalanced).
    """


    def __init__(self, agent: Agent, initial_condition: np.ndarray, enable_log: bool = False):
        super(DirectedGradientTracking, self).__init__(agent, initial_condition, enable_log)

        # initialize tracker
        self.y = self.agent.problem.objective_function.subgradient(initial_condition)


    def iterate_run(self, **kwargs):
        """Run a single iterate of the algorithm
        """
        stepsize = kwargs.get('stepsize', 0.1)
        # in
        data = self.agent.neighbors_exchange(self.x)

        for neigh in data:
            self.x_neigh[neigh] = data[neigh]

        x_avg = self.agent.in_weights[self.agent.id] * self.x
        for i in self.agent.in_neighbors:
            x_avg += self.agent.in_weights[i] * self.x_neigh[i]

        subg_old = self.agent.problem.objective_function.subgradient(self.x)
        self._update_x_average(x_avg - stepsize * self.y, **kwargs)
        subg_new = self.agent.problem.objective_function.subgradient(self.x)

        # y average
        send_data = {}
        gradient_diff = subg_new - subg_old
        for j in self.agent.out_neighbors:
            send_data[j] = self.agent.out_weights[j] * (self.y + gradient_diff)
        data = self.agent.neighbors_exchange(send_data, dict_neigh=True)

        for neigh in data:
            self.y_neigh[neigh] = data[neigh]

        y_avg = self.agent.out_weights[self.agent.id] * (self.y + gradient_diff)
        for i in self.agent.in_neighbors:
            y_avg += self.y_neigh[i]

        self._update_y_average(y_avg, **kwargs)

    def get_result(self):
        """Return the actual value of x

        Returns:
            numpy.ndarray: value of x
        """
        return self.x



import setuptools

with open("README_pkg.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="disropt",
    version="0.1.8",
    author="Francesco Farina, Andrea Camisa, Andrea Testa, Ivano Notarnicola, Giuseppe Notarstefano",
    author_email="franc.farina@unibo.it",
    description="DISROPT: a python framework for distributed optimization",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://opt4smart.github.io/disropt/",
    packages=setuptools.find_packages(),
    install_requires=[
        'autograd>=1.3',
        'cvxopt>=1.2.3',
        'cvxpy>=1.0.25',
        'mpi4py>=3.0.1',
        'numpy>=1.16.3',
        'osqp>=0.6.1',
        'scipy>=1.2.1',
        'dill>=0.3.0'
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)


import numpy as np
import copy
from .abstract_function import AbstractFunction


class SubmodularFn(AbstractFunction):
    """Submodular AbstractFunction abstract class

    Attributes:
        V: ground set
        input_shape: cardinality of the ground set

    Args:
        input_shape: cardinality of the ground set

    Raises:
        ValueError: input_shape must be an integer positive number
    """

    def __init__(self, input_shape):

        if not isinstance(input_shape, int) or input_shape < 0:
            raise ValueError(
                "Input must be a positive integer")

        self.V = np.arange(input_shape)
        self.input_shape = (input_shape, 1)

    def eval(self, set):
        """
        Evaluate the submodular function at a given set

        Parameters:
                set (numpy.ndarray(, dtype='int')): vector

                """
        pass

    def greedy_polyhedron(self, w):
        """Greedy algorithm for finding a maximizer :math:`x` of

        .. math::
            :label: basepoly

                \\max_{x\\in B(F)}    & w^\\top x \\\\


        where :math:`B(F)` is the base polyhedron associated to the submodular function :math:`F`

        Parameters:
                w: cost direction

        Returns:
            numpy.ndarray: maximizer of :eq:`basepoly`

        Raises:
            ValueError: Input must be a numpy.ndarray with input_shape elements
                """

        w = w.flatten()
        n = len(self.V)
        Ind = np.flip(np.argsort(w))
        w = w[Ind]
        x = np.zeros((n, 1))
        empt = np.array([], dtype='int')
        Fold = self.eval(empt[:, None])
        A = np.ones((1, 1), dtype='int')
        A[0, :] = self.V[Ind[0]]
        x[Ind[0]] = self.eval(A) - Fold
        Fold = Fold + x[Ind[0]]

        for i in range(1, n):
            Anew = np.vstack((A, self.V[Ind[i]]))
            x[Ind[i]] = self.eval(Anew) - Fold
            A = Anew
            Fold = Fold + x[Ind[i]]
        return x

    def subgradient(self, x):
        """
        Evaluate a subgradient of the Lovasz extension of the
        submodular function at :math:`x`

        Parameters:
                x: vector
        Returns:
            numpy.ndarray: subgradient of the Lovasz extension of :math:`F` at :math:`x`

        Raises:
            ValueError: Input must be a numpy.ndarray with input_shape elements
                """
        return self.greedy_polyhedron(x)

    def blocksubgradient(self, x, block):
        """
        Evaluate a subgradient of the Lovasz extension of the
        submodular function at :math:`x`

        Parameters:
                x: vector
                block: vector
        Returns:
            numpy.ndarray: block subgradient of the Lovasz extension of :math:`F` at :math:`x`

        Raises:
            ValueError: Input must be a numpy.ndarray with input_shape elements
                """
        xb = copy.deepcopy(x)
        xb = xb.flatten()
        n = len(self.V)

        idx = np.where(xb[block] == np.amin(xb[block]))
        idx = np.asarray(idx)
        right_index = block[idx.item(0)]
        pivot = xb[right_index]
        rg_idx = np.arange(n)
        xb[0], xb[right_index] = xb[right_index], xb[0]
        rg_idx[0], rg_idx[right_index] = rg_idx[right_index], rg_idx[0]
        i = 1
        for j in np.arange(1, len(xb)):
            if xb[j] >= pivot:
                xb[j], xb[i] = xb[i], xb[j]
                rg_idx[j], rg_idx[i] = rg_idx[i], rg_idx[j]
                i += 1
        xb[0], xb[i - 1] = xb[i - 1], xb[0]
        rg_idx[0], rg_idx[i-1] = rg_idx[i-1], rg_idx[0]

        i_n = np.where(rg_idx == right_index)
        i_n = i_n[0].item(0)+1
        Ind = np.flip(np.argsort(xb[:i_n]))
        Ind = rg_idx[:i_n][Ind]
        lb = len(block)
        count = 1
        y = np.zeros((n, 1))

        empt = np.array([], dtype='int')
        Fold = self.eval(empt[:, None])
        A = np.ones((1, 1), dtype='int')
        A[0, :] = self.V[Ind[0]]
        prev = -1
        i = 1
        if Ind[0] in block:
            y[Ind[0]] = self.eval(A) - Fold
            Fold = Fold + y[Ind[0]]
            count += 1
            prev = 0
        while count <= lb:
            Anew = np.vstack((A, self.V[Ind[i]]))
            if Ind[i] in block:
                Fnew = self.eval(Anew)
                if i == prev+1:
                    y[Ind[i]] = Fnew - Fold
                else:
                    y[Ind[i]] = Fnew - self.eval(A)
                prev = i
                Fold = Fnew
                count += 1
            A = Anew
            i += 1
        return y


class stCutFn(SubmodularFn):
    """Submodular AbstractFunction for evaluating s-t cuts
       in a network of N+2 nodes where
       s node is the N+1 - th node and
       t node is the N+2 - th node

    Attributes:
        G: (N+2)x(N+2) communication graph
        eval_cut (numpy.ndarray,numpy.ndarray): evaluate cut value.
        eval (int): wrapper for eval_cut

    Args:
        input_shape (int): ground set cardinality
        G (numpy.ndarray): (input_shape+2)x(input_shape+2) matrix

    Raises:
        ValueError: input_shape must be an integer positive number
        ValueError: G must be a (input_shape+2)x(input_shape+2) numpy.ndarray

    """

    def __init__(self, input_shape, G):
        """Initialize the function

        """
        super(stCutFn, self).__init__(input_shape)

        if not isinstance(G, np.ndarray) or not (G.shape[0] == G.shape[1]) or not (G.shape[0] == (input_shape + 2)):
            raise ValueError(
                "Input must be a numpy.ndarray with shape {},{}".format(
                    input_shape + 2, input_shape + 2))

        self.G = G

    def eval_cut(self, G, A):
        """Evaluate cut value

        Args:
            G (numpy.ndarray): network graph
            A (numpy.ndarray): set of nodes

        Returns:
            float: cut value

        Raises:
            ValueError: A must be a numpy.ndarray with size at most (1,input_shape)

        """
        if not isinstance(
            A, np.ndarray) or (
            A.shape[1] > self.input_shape[1]) or (
                A.shape[1] > 1):
            raise ValueError(
                "Input must be a numpy.ndarray with shape at most {}".format(self.input_shape[0]))

        n = G.shape[0]
        A = np.vstack((A, n-2))  # Add s to set
        G1 = G[:, np.setdiff1d(np.arange(n), A)]
        G2 = G[:, np.setdiff1d(np.arange(n), [n - 2])]
        C = np.sum(G1[A]) - np.sum(G2[n - 2, :])
        return C

    def eval(self, A):
        """wrapper for eval cut function

        Args:
            A (numpy.ndarray): set of nodes

        Returns:
            float: cut value

        Raises:
            ValueError: A must be a numpy.ndarray with size at most (|V|, 1)
        """
        # TODO is instace np.ndarray(int)
        if not isinstance(A, np.ndarray) or (A.shape[0] > self.input_shape[0]) or (A.shape[1] > 1):
            raise ValueError(
                "Input must be a numpy.ndarray(dtype=int) with shape at most {}".format(self.input_shape[0]))
        C = self.eval_cut(self.G, A)

        return C

class Cardinality(SubmodularFn):

    def eval(self, A):
        """

        Args:
            A (numpy.ndarray): set

        Returns:
            float: cardinality of A

        Raises:
            ValueError: A must be a numpy.ndarray with size at most (|V|, 1)
        """
        # TODO is instace np.ndarray(int)
        if not isinstance(A, np.ndarray) or (A.shape[0] > self.input_shape[0]) or (A.shape[1] > 1):
            raise ValueError(
                "Input must be a numpy.ndarray(dtype=int) with shape at most {}".format(self.input_shape[0]))
        return len(A)


from ..agents import Agent


class Algorithm:
    """Algorithm abstract class

    Args:
        agent (Agent): agent to execute the algorithm
        enable_log (bool): True for enabling log

    Attributes:
        agent (Agent): agent to execute the algorithm
        sequence (numpy.ndarray): sequence of data generated by the algorithm
        enable_log (bool): True for enabling log
    """

    def __init__(self, agent: Agent, enable_log: bool=False, **kwargs):
        if not isinstance(agent, Agent):
            raise ValueError("agent must be an Agent object")
        if not isinstance(enable_log, bool):
            raise ValueError("enable_log must be a bool")
        self.agent = agent
        self.enable_log = enable_log
        self.sequence = None
        super().__init__(**kwargs)

    def run(self):
        """Run the algorithm
        """
        pass

    def get_result(self):
        """Return the value of the solution
        """
        pass



import numpy as np
from typing import Union, Callable, List
from threading import Event
from ..agents.agent import Agent
from ..problems import ConvexifiedMILP, ConstraintCoupledMILP, MixedIntegerLinearProblem
from .primal_decomp import PrimalDecomposition
from ..functions import Variable
from . import MaxConsensus


class PrimalDecompositionMILP(PrimalDecomposition):
    """Distributed primal decomposition for MILPs.
    """

    # TODO choose ref
    def __init__(self, agent: Agent, initial_condition: np.ndarray, enable_log: bool = False,
        restriction: np.ndarray = None, finite_time_add_restriction: float = 0):

        super().__init__(agent, initial_condition, enable_log)

        # check input
        if not isinstance(agent.problem, ConstraintCoupledMILP):
            raise TypeError("The agent must be equipped with a ConstraintCoupledMILP")

        # initialize some variables and local MILP solver
        self.fast_mode = True
        param = agent.problem.coupling_function.get_parameters()
        self.A = param[0].T
        self.cutting_planes = None
        self.restriction = restriction
        self.ft_restriction = finite_time_add_restriction
        self.integer_vars = agent.problem.integer_vars
        self.binary_vars = agent.problem.binary_vars
        self.milp = ConvexifiedMILP(objective_function = agent.problem.objective_function,
            y = np.zeros((self.S, 1)), A = self.A, constraints = agent.problem.constraints,
            integer_vars = self.integer_vars, binary_vars = self.binary_vars)
        
        # further parameters of local solver
        self.max_local_iterations = 1000
        self.local_threshold_convergence = 1e-8
        self.viol_additional_slack = 1e-5 # additional slack in case of violation

    def run(self, n_agents: int, iterations: int = 1000, stepsize: Union[float, Callable] = 0.1, M: float = 1000.0,
        verbose: bool=False, callback_iter: Callable=None, fast_mode: bool=True,
        max_cutting_planes: int=1000, milp_solver: str=None,
        extra_allocation: np.ndarray = None,
        use_runavg: bool=False, # use running average of y in local MILP?
        runavg_start_iter: int=0, event: Event=None,
        max_consensus_iterations: int = None,
        max_consensus_graph_diam: int = None, **kwargs) -> np.ndarray:
        """Run the algorithm for a given number of iterations

        Args:
            iterations: Number of iterations. Defaults to 1000.
            stepsize: If a float is given as input, the stepsize is constant. 
                      If a function is given, it must take an iteration k as input and output the corresponding stepsize. Defaults to 0.1.
            M: Value of the parameter :math:`M`. Defaults to 1000.
            verbose: If True print some information during the evolution of the algorithm. Defaults to False.
            callback_iter: callback function to be called at the end of each iteration. Must take an iteration k as input. Defaults to None.
            fast_mode: If True, a mixed-integer solution is computed at each iteration, otherwise only at the last iteration. Defaults to True.
            max_cutting_planes: maximum number of cutting planes for local solver. Defaults to 1000.
            milp_solver: MILP solver to use. Defaults to None (use default solver).

        Raises:
            TypeError: The number of iterations must be an int
            TypeError: The stepsize must be a float or a callable
            TypeError: The parameter M must be a float

        Returns:
            return a tuple (x, y) with the sequence of primal solutions and allocation estimates if enable_log=True.
            If fast_mode=True, the primal solutions are those of the convexified problem, otherwise they are the mixed-integer estimates.
        """

        # store parameters
        self.milp_solver = milp_solver
        self.max_cutting_planes = max_cutting_planes
        self.fast_mode = fast_mode
        self.extra_allocation = extra_allocation if extra_allocation is not None else np.zeros((self.S, 1))
        self.use_runavg = use_runavg

        # compute and apply restriction
        self.compute_restriction(max_consensus_iterations, max_consensus_graph_diam)
        self.y = self.y0 - (self.restriction + self.ft_restriction)/n_agents

        if verbose and self.agent.id == 0:
            total_restriction = np.array(self.restriction + self.ft_restriction)
            print('Overall applied restriction: {}'.format(total_restriction.flatten()))

        # run primal decomposition algorithm on convexified problem
        result = super().run(iterations, stepsize, M, verbose, callback_iter, compute_runavg=use_runavg, runavg_start_iter=runavg_start_iter, event=event, **kwargs)

        # compute mixed-integer solution (if not already available)
        if fast_mode and (event is None or not event.is_set()): # skip if optimization has been interrupted
            self.x = self.mixed_integer_solution()
        
        return result
    
    def iterate_run(self, stepsize: float, M: float, update_runavg, event: Event, **kwargs):
        """Run a single iterate of the algorithm
        """

        # solve local problem
        out = self.milp.solve(M=M, milp_solver=self.milp_solver, return_only_solution=False,
                y=self.y, cutting_planes=self.cutting_planes, max_iterations=self.max_local_iterations,
                threshold_convergence=self.local_threshold_convergence)

        # save data
        x = out[0][0]
        mu = out[2]
        self.cutting_planes = out[3]

        # exchange dual variables with neighbors
        data = self.agent.neighbors_exchange(mu, event=event)
        mu_neigh = [data[idx] for idx in data]

        if event is None or not event.is_set():
            self._update_local_solution(x, mu, mu_neigh, stepsize, update_runavg, **kwargs)

            # solve local MILP if requested
            if not self.fast_mode:
                self.x = self.mixed_integer_solution()
    
    def compute_restriction(self, iterations, graph_diam):
        # check if restriction is already available
        if self.restriction is not None:
            return
        
        # compute lower and upper bounds of allocations
        LB = np.zeros((self.S, 1))
        UB = np.zeros((self.S, 1))

        for s in range(self.S):
            x = Variable(self.x_shape[0])
            obj_func = self.A[s, :][:, None] @ x

            # solve s-th lower bound problem
            milp = MixedIntegerLinearProblem(obj_func, self.agent.problem.constraints, self.integer_vars, self.binary_vars)
            x_opt = milp.solve() if self.milp_solver is None else milp.solve(solver=self.milp_solver)
            
            LB[s] = obj_func.eval(x_opt)

            # solve s-th upper bound problem
            milp.objective_function = -obj_func
            x_opt = milp.solve() if self.milp_solver is None else milp.solve(solver=self.milp_solver)
            
            UB[s] = obj_func.eval(x_opt)

        # prepare minimum resource MILP
        A = np.hstack((np.zeros((self.S, self.x_shape[0])), np.ones((self.S,1)))).transpose() # matrix for \rho \1
        A_rho = np.hstack((np.zeros((1, self.x_shape[0])), [[1]])).transpose() # matrix selecting only rho variable

        z = Variable(self.x_shape[0] + 1) # symbolic variable [x, rho]
        rho = A_rho @ z # symbolic variable (only rho)

        constraints = [self.coupling_function <= LB + A @ z]
        constraints.extend(self.local_constraints)

        # solve MILP to compute vector using minimal resources
        milp = MixedIntegerLinearProblem(rho, constraints, self.integer_vars, self.binary_vars)
        z_opt = milp.solve() if self.milp_solver is None else milp.solve(solver=self.milp_solver)

        # compute worst-case local violation (finally!)
        min_resource = rho.eval(z_opt) * np.ones((self.S, 1))
        restriction_loc = np.minimum(UB - LB, min_resource)

        ######################

        # run max-consensus algorithm
        algorithm = MaxConsensus(self.agent, restriction_loc, graph_diam)

        if iterations is not None:
            algorithm.run(iterations=iterations)
        else:
            algorithm.run()

        # save result
        self.restriction = self.S * algorithm.get_result()
    
    def mixed_integer_solution(self):

        allocation = self.y_avg if self.use_runavg else self.y

        y_viol = self.compute_violating_y(allocation)

        # prepare final MILP
        x = Variable(self.x_shape[0])
        constraints = [self.A.T@x <= y_viol + self.extra_allocation]
        constraints.extend(self.agent.problem.constraints)

        # solve final MILP
        milp_final = MixedIntegerLinearProblem(self.agent.problem.objective_function, constraints,
            self.integer_vars, self.binary_vars)
        x_final = milp_final.solve() if self.milp_solver is None else milp_final.solve(solver=self.milp_solver)

        return x_final
    
    def compute_violating_y(self, allocation):

        A = np.hstack((np.zeros((self.S, self.x_shape[0])), np.ones((self.S,1)))).transpose() # matrix for \rho \1
        A_rho = np.hstack((np.zeros((1, self.x_shape[0])), [[1]])).transpose() # matrix selecting only rho variable

        z = Variable(self.x_shape[0] + 1) # symbolic variable [x, rho]
        rho = A_rho @ z # symbolic variable (only rho)
        alloc_constr = self.coupling_function <= allocation + self.extra_allocation + A @ z
        constraints = [alloc_constr, rho >= 0]
        constraints.extend(self.local_constraints)

        # solve MILP to compute violation
        milp_viol = MixedIntegerLinearProblem(rho, constraints, self.integer_vars, self.binary_vars)
        z_opt = milp_viol.solve() if self.milp_solver is None else milp_viol.solve(solver=self.milp_solver)

        viol = rho.eval(z_opt) # evaluate rho at optimal solution

        # vectorize violation
        violation = (viol + self.viol_additional_slack) * np.ones((self.S, 1))

        return allocation + violation



import numpy as np
import autograd
import autograd.numpy as anp
from functools import wraps  # AT: I added this line


# decorator for checking inputs to eval and subgradient


def check_input(checked_fn):
    @wraps(checked_fn)  # AT: I added this line
    def check(fn, x, **kwargs):
        if x is not None:
            if not isinstance(x, (np.ndarray, anp.numpy_boxes.ArrayBox)) or not (fn.input_shape == x.shape):
                raise ValueError(
                    "Input must be a numpy.ndarray with shape {}".format(
                        fn.input_shape))
        block = kwargs.get("block", None)
        if block is not None:
            if not isinstance(block, np.ndarray):
                raise TypeError("block argument must be a numpy.ndarray")
        return checked_fn(fn, x, **kwargs)
    return check



import numpy as np


def is_pos_def(P: np.ndarray) -> bool:
    """check if a matrix is positive definite
    
    Args:
        P (numpy.ndarray): matrix
    
    Returns:
        bool: 
    """
    return np.all(np.linalg.eigvals(P) > 0)

def is_semi_pos_def(P: np.ndarray) -> bool:
    """check if a matrix is positive semi-definite
    
    Args:
        P (numpy.ndarray): matrix
    
    Returns:
        bool: 
    """
    return np.all(np.linalg.eigvals(P) >= 0)

def check_symmetric(A: np.ndarray, rtol: float=1e-05, atol: float=1e-08) -> bool:
    """check if a matrix is symetric
    
    Args:
        A (numpy.ndarray): matrix
        rtol (float): Defaults to 1e-05.
        atol (float): Defaults to 1e-08.
    
    Returns:
        bool: 
    """
    return np.allclose(A, A.transpose(), rtol=rtol, atol=atol)


from ..constraints import Constraint, AbstractSet

def check_affine_constraints(constraints: list) -> bool:
    """Check if a list of Constraint objects contains only affine constraints
    
    Args:
        constraints (list): list of Constraint
    
    Returns:
        bool
    """
    affine_constraints = True
    if constraints is not None:
        if not isinstance(constraints, list):
            constraints = [constraints]
        for constraint in constraints:
            if isinstance(constraint, Constraint):
                if not constraint.is_affine:
                    affine_constraints = False
            if isinstance(constraint, AbstractSet):
                constraints_list = constraint.to_constraints()
                for cns in constraints_list:
                    if not cns.is_affine:
                        affine_constraints = False
    return affine_constraints
